# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'functions_manager.ui'
##
## Created by: Qt User Interface Compiler version 6.3.1
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QGridLayout, QGroupBox, QHBoxLayout,
    QListWidgetItem, QPlainTextEdit, QPushButton, QSizePolicy,
    QSpacerItem, QWidget)

from asammdf.gui.widgets.list import MinimalListWidget
from . import resource_rc

class Ui_FunctionsManager(object):
    def setupUi(self, FunctionsManager):
        if not FunctionsManager.objectName():
            FunctionsManager.setObjectName(u"FunctionsManager")
        FunctionsManager.resize(749, 703)
        self.gridLayout_2 = QGridLayout(FunctionsManager)
        self.gridLayout_2.setObjectName(u"gridLayout_2")
        self.horizontalLayout_2 = QHBoxLayout()
        self.horizontalLayout_2.setSpacing(10)
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.add_btn = QPushButton(FunctionsManager)
        self.add_btn.setObjectName(u"add_btn")
        icon = QIcon()
        icon.addFile(u":/plus.png", QSize(), QIcon.Normal, QIcon.Off)
        self.add_btn.setIcon(icon)
        self.add_btn.setIconSize(QSize(24, 24))

        self.horizontalLayout_2.addWidget(self.add_btn)

        self.erase_btn = QPushButton(FunctionsManager)
        self.erase_btn.setObjectName(u"erase_btn")
        icon1 = QIcon()
        icon1.addFile(u":/erase.png", QSize(), QIcon.Normal, QIcon.Off)
        self.erase_btn.setIcon(icon1)
        self.erase_btn.setIconSize(QSize(24, 24))

        self.horizontalLayout_2.addWidget(self.erase_btn)

        self.import_btn = QPushButton(FunctionsManager)
        self.import_btn.setObjectName(u"import_btn")
        icon2 = QIcon()
        icon2.addFile(u":/open.png", QSize(), QIcon.Normal, QIcon.Off)
        self.import_btn.setIcon(icon2)
        self.import_btn.setIconSize(QSize(24, 24))

        self.horizontalLayout_2.addWidget(self.import_btn)

        self.export_btn = QPushButton(FunctionsManager)
        self.export_btn.setObjectName(u"export_btn")
        icon3 = QIcon()
        icon3.addFile(u":/save.png", QSize(), QIcon.Normal, QIcon.Off)
        self.export_btn.setIcon(icon3)
        self.export_btn.setIconSize(QSize(24, 24))

        self.horizontalLayout_2.addWidget(self.export_btn)

        self.horizontalSpacer_5 = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.horizontalLayout_2.addItem(self.horizontalSpacer_5)


        self.gridLayout_2.addLayout(self.horizontalLayout_2, 0, 0, 1, 1)

        self.function_group = QGroupBox(FunctionsManager)
        self.function_group.setObjectName(u"function_group")
        self.gridLayout = QGridLayout(self.function_group)
        self.gridLayout.setObjectName(u"gridLayout")
        self.horizontalLayout_3 = QHBoxLayout()
        self.horizontalLayout_3.setObjectName(u"horizontalLayout_3")
        self.check_syntax_btn = QPushButton(self.function_group)
        self.check_syntax_btn.setObjectName(u"check_syntax_btn")
        icon4 = QIcon()
        icon4.addFile(u":/checkmark.png", QSize(), QIcon.Normal, QIcon.Off)
        self.check_syntax_btn.setIcon(icon4)

        self.horizontalLayout_3.addWidget(self.check_syntax_btn)

        self.store_btn = QPushButton(self.function_group)
        self.store_btn.setObjectName(u"store_btn")
        icon5 = QIcon()
        icon5.addFile(u":/shift_down.png", QSize(), QIcon.Normal, QIcon.Off)
        self.store_btn.setIcon(icon5)

        self.horizontalLayout_3.addWidget(self.store_btn)

        self.horizontalSpacer_2 = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.horizontalLayout_3.addItem(self.horizontalSpacer_2)


        self.gridLayout.addLayout(self.horizontalLayout_3, 1, 0, 1, 2)

        self.function_definition = QPlainTextEdit(self.function_group)
        self.function_definition.setObjectName(u"function_definition")
        font = QFont()
        font.setFamilies([u"Consolas"])
        self.function_definition.setFont(font)
        self.function_definition.setHorizontalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        self.function_definition.setLineWrapMode(QPlainTextEdit.NoWrap)
        self.function_definition.setTabStopDistance(140.000000000000000)

        self.gridLayout.addWidget(self.function_definition, 0, 0, 1, 2)

        self.gridLayout.setRowStretch(0, 1)

        self.gridLayout_2.addWidget(self.function_group, 0, 1, 2, 1)

        self.functions_list = MinimalListWidget(FunctionsManager)
        self.functions_list.setObjectName(u"functions_list")
        self.functions_list.setMinimumSize(QSize(300, 0))

        self.gridLayout_2.addWidget(self.functions_list, 1, 0, 1, 1)

        self.gridLayout_2.setColumnStretch(0, 1)
        self.gridLayout_2.setColumnStretch(1, 2)

        self.retranslateUi(FunctionsManager)

        QMetaObject.connectSlotsByName(FunctionsManager)
    # setupUi

    def retranslateUi(self, FunctionsManager):
        FunctionsManager.setWindowTitle(QCoreApplication.translate("FunctionsManager", u"Form", None))
        self.add_btn.setText(QCoreApplication.translate("FunctionsManager", u"Add", None))
#if QT_CONFIG(tooltip)
        self.erase_btn.setToolTip("")
#endif // QT_CONFIG(tooltip)
        self.erase_btn.setText(QCoreApplication.translate("FunctionsManager", u"Delete all", None))
        self.import_btn.setText(QCoreApplication.translate("FunctionsManager", u"Load definitions", None))
        self.export_btn.setText(QCoreApplication.translate("FunctionsManager", u"Save definitions", None))
        self.function_group.setTitle(QCoreApplication.translate("FunctionsManager", u"Function definition", None))
        self.check_syntax_btn.setText(QCoreApplication.translate("FunctionsManager", u"Check syntax", None))
        self.store_btn.setText(QCoreApplication.translate("FunctionsManager", u"Store function changes", None))
        self.function_definition.setPlaceholderText(QCoreApplication.translate("FunctionsManager", u"function code", None))
    # retranslateUi




# -*- coding: utf-8 -*-
import inspect
import json
import math
import os
from pathlib import Path
import random
from traceback import format_exc

from natsort import natsorted
import numpy as np
import pandas as pd
from PySide6 import QtCore, QtGui, QtWidgets

from ..dialogs.messagebox import MessageBox
from ..ui.functions_manager import Ui_FunctionsManager
from ..utils import (
    check_generated_function,
    ErrorDialog,
    FUNC_NAME,
    generate_python_function,
)
from .python_highlighter import PythonHighlighter


class FunctionsManager(Ui_FunctionsManager, QtWidgets.QWidget):
    def __init__(
        self, definitions, channels=None, selected_definition="", *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.setupUi(self)

        self.channels = channels or {}

        self.definitions = definitions

        self.function_definition.setPlaceholderText(
            """The virtual channel definition is written as a Python function.
Use the 'return' statement to return a value, otherwise 'None' will automatically be returned by the function.
The last function argument must be 't=0'.

Here is a minimalistic example:

def MyAverage(main_clock=0, p_FL=0, p_FR=0, p_RL=0, p_RR=0, vehicle_speed=0, t=0): 
    if main_clock > 5000:
        return 0
    else:
        avg = (p_FL + p_FR + p_RL + p_RR) / 4
        
        if avg > 20.5 and vehicle_speed < 100:
            return avg
        else:
            return avg + 9
"""
        )
        self.function_definition.setTabStopDistance(
            QtGui.QFontMetricsF(self.function_definition.font()).horizontalAdvance(" ")
            * 4
        )

        p = self.function_definition.palette()
        for active in (QtGui.QPalette.Active, QtGui.QPalette.Inactive):
            p.setColor(active, QtGui.QPalette.Base, QtGui.QColor("#131314"))
            p.setColor(active, QtGui.QPalette.WindowText, QtGui.QColor("#ffffff"))
            p.setColor(active, QtGui.QPalette.Text, QtGui.QColor("#ffffff"))
        self.function_definition.setPalette(p)

        PythonHighlighter(self.function_definition.document())

        self.functions_list.minimal_menu = True
        self.functions_list.all_texts = True
        self.functions_list.placeholder_text = (
            "Press the + button to add a new function definition"
        )
        self.functions_list.user_editable = True
        self.functions_list.setAlternatingRowColors(True)

        self.functions_list.currentItemChanged.connect(
            self.definition_selection_changed
        )
        self.functions_list.itemsDeleted.connect(self.definitions_deleted)

        self.add_btn.clicked.connect(self.add_definition)
        self.check_syntax_btn.clicked.connect(self.check_syntax)
        self.erase_btn.clicked.connect(self.erase_definitions)
        self.export_btn.clicked.connect(self.export_definitions)
        self.import_btn.clicked.connect(self.import_definitions)
        self.store_btn.clicked.connect(self.store_definition)

        if self.definitions:
            self.functions_list.clear()
            names = natsorted(self.definitions)
            self.functions_list.addItems(names)
            if selected_definition in names:
                self.functions_list.setCurrentRow(names.index(selected_definition))
            elif names:
                self.functions_list.setCurrentRow(0)

        for button in (
            self.add_btn,
            self.check_syntax_btn,
            self.erase_btn,
            self.import_btn,
            self.export_btn,
        ):
            button.setDefault(False)
            button.setAutoDefault(False)

        self.showMaximized()

    def add_definition(self):
        counter = 1
        while True:
            name = f"Function{counter}"
            if name in self.definitions:
                counter += 1
            else:
                break

        self.definitions[name] = {
            "definition": f"def {name}(t=0):\n    return 0",
            "uuid": os.urandom(6).hex(),
        }

        self.functions_list.clear()
        names = natsorted(self.definitions)
        self.functions_list.addItems(names)

        row = names.index(name)
        self.functions_list.setCurrentRow(row)

    def check_syntax(self, silent=False):
        _globals = {
            "math": math,
            "np": np,
            "pd": pd,
        }

        for info in self.definitions.values():
            generate_python_function(info["definition"], in_globals=_globals)

        function_source = self.function_definition.toPlainText().replace("\t", "    ")
        func, trace = generate_python_function(function_source, in_globals=_globals)

        return check_generated_function(
            func, trace, function_source, silent, parent=self
        )

    def definitions_deleted(self, deleted):
        count = self.functions_list.count()
        names = set(self.functions_list.item(row).text() for row in range(count))

        deleted = [name for name in self.definitions if name not in names]

        for name in deleted:
            del self.definitions[name]

        if not self.definitions:
            self.function_definition.setPlainText("")

    def definition_selection_changed(self, current, previous):
        if current:
            name = current.text()

            definition = self.definitions[name]
            self.function_definition.setPlainText(definition["definition"])

    def erase_definitions(self):
        self.functions_list.clear()
        self.definitions = {}
        self.function_definition.setPlainText("")

    def export_definitions(self, *args):
        self.refresh_definitions()

        file_name, _ = QtWidgets.QFileDialog.getSaveFileName(
            self,
            "Select functions definition export file",
            "",
            "Functions definition file (*.def)",
        )

        if file_name and Path(file_name).suffix.lower() == ".def":
            definitions = {
                name: info["definition"] for name, info in self.definitions.items()
            }
            Path(file_name).write_text(json.dumps(definitions, indent=2))

    def import_definitions(self, *args):
        self.refresh_definitions()

        file_name, _ = QtWidgets.QFileDialog.getOpenFileName(
            self,
            "Select functions definition file",
            "",
            "Functions definitions file (*.def)",
            "Functions definitions file (*.def)",
        )

        if file_name and Path(file_name).suffix.lower() == ".def":
            file_name = Path(file_name)

            with open(file_name, "r") as infile:
                info = json.load(infile)

            for name, definition in info.items():
                if name in self.definitions:
                    self.definitions[name]["definition"] = definition
                else:
                    self.definitions[name] = {
                        "definition": definition,
                        "uuid": os.urandom(6).hex(),
                    }

            self.functions_list.clear()
            names = natsorted(self.definitions)
            self.functions_list.addItems(names)
            if names:
                self.functions_list.setCurrentRow(0)

    def refresh_definitions(self):
        _globals = {
            "math": math,
            "np": np,
            "pd": pd,
        }

        for info in self.definitions.values():
            generate_python_function(info["definition"], in_globals=_globals)

        function_source = self.function_definition.toPlainText().replace("\t", "    ")
        func, trace = generate_python_function(function_source, in_globals=_globals)

        if func is not None:
            name = func.__name__
            if name in self.definitions:
                self.definitions[name][
                    "definition"
                ] = self.function_definition.toPlainText()

    def store_definition(self, *args):
        ok, func = self.check_syntax(silent=True)
        if ok:
            item = self.functions_list.currentItem()
            name = func.__name__

            if name != item.text() and name in self.definitions:
                MessageBox.information(
                    self,
                    "Invalid function name",
                    f'The name "{name}" is already given to another function.\n'
                    "The function names must be unique",
                )
                return

            info = self.definitions.pop(item.text())
            info["definition"] = self.function_definition.toPlainText()
            self.definitions[func.__name__] = info

            self.functions_list.clear()
            names = natsorted(self.definitions)
            self.functions_list.addItems(names)

            row = names.index(name)
            self.functions_list.setCurrentRow(row)



# -*- coding: utf-8 -*-

from datetime import datetime, timedelta, timezone

from PySide6 import QtCore, QtGui, QtWidgets

LOCAL_TIMEZONE = datetime.now(timezone.utc).astimezone().tzinfo

import numpy as np
import pandas as pd
import pyqtgraph as pg
import pyqtgraph.functions as fn
from pyqtgraph.graphicsItems.ButtonItem import ButtonItem
from pyqtgraph.Point import Point

BUTTON_SIZE = 16


class FormatedAxis(pg.AxisItem):
    rangeChanged = QtCore.Signal(object, object)
    scale_editor_requested = QtCore.Signal(object)

    def __init__(self, *args, **kwargs):
        self.plus = self.minus = None
        self.uuid = kwargs.pop("uuid", None)
        self.background = kwargs.pop("background", fn.mkColor("#000000"))

        super().__init__(*args, **kwargs)

        self.format = "phys"
        self.mode = "phys"
        self.text_conversion = None
        self.origin = None

        self._label_with_unit = ""

        self.locked = kwargs.pop("locked", False)

        if self.orientation in ("left", "right"):
            self.plus = ButtonItem(":/plus.png", BUTTON_SIZE, parentItem=self)
            self.minus = ButtonItem(":/minus.png", BUTTON_SIZE, parentItem=self)

            self.plus.clicked.connect(self.increase_width)
            self.minus.clicked.connect(self.decrease_width)

            if self.scene() is not None:
                self.scene().addItem(self.plus)
                self.scene().addItem(self.minus)

            self.setWidth(48)

        self.set_pen(self._pen)

    def increase_width(self):
        self.setWidth(self.width() + 10)

    def decrease_width(self):
        self.setWidth(self.width() - 10)

    def setWidth(self, w=None):
        if self.orientation in ("left", "right"):
            super().setWidth(max(w, 48))
        else:
            super().setWidth(w)

    def tickStrings(self, values, scale, spacing):
        strns = []
        self.tick_positions = values

        if self.text_conversion and self.mode == "phys":
            strns = []
            for val in values:
                nv = self.text_conversion.convert(np.array([val]), as_bytes=True)[0]

                val = float(val)

                if val.is_integer():
                    val = int(val)

                    if self.format == "hex":
                        val = hex(int(val))
                    elif self.format == "bin":
                        val = bin(int(val))
                    else:
                        val = str(val)
                else:
                    val = f"{val:.6f}"

                if isinstance(nv, bytes):
                    try:
                        strns.append(f'{val}={nv.decode("utf-8", errors="replace")}')
                    except:
                        strns.append(f'{val}={nv.decode("latin-1", errors="replace")}')
                else:
                    strns.append(val)
        else:
            if self.format == "phys":
                strns = super(FormatedAxis, self).tickStrings(values, scale, spacing)

            elif self.format == "hex":
                for val in values:
                    val = float(val)
                    if val.is_integer():
                        val = hex(int(val))
                    else:
                        val = ""
                    strns.append(val)

            elif self.format == "bin":
                for val in values:
                    val = float(val)
                    if val.is_integer():
                        val = bin(int(val))
                    else:
                        val = ""
                    strns.append(val)
            elif self.format == "ascii":
                for val in values:
                    val = float(val)
                    if val.is_integer():
                        val = int(val)
                        if 0 < val < 0x110000:
                            val = chr(val)
                        else:
                            val = str(val)
                    else:
                        val = ""
                    strns.append(val)
            elif self.format == "time":
                strns = [str(timedelta(seconds=val)) for val in values]
            elif self.format == "date":
                strns = (
                    pd.to_datetime(
                        np.array(values) + self.origin.timestamp(),
                        unit="s",
                        errors="coerce",
                    )
                    .tz_localize("UTC")
                    .tz_convert(LOCAL_TIMEZONE)
                    .astype(str)
                    .to_list()
                )

        return [val[:80] for val in strns]

    def setLabel(self, text=None, units=None, unitPrefix=None, **args):
        """overwrites pyqtgraph setLabel"""
        show_label = False
        if text is not None:
            self.labelText = text
            show_label = True
        if units is not None:
            self.labelUnits = units
            show_label = True
        if show_label:
            self.showLabel()
        if unitPrefix is not None:
            self.labelUnitPrefix = unitPrefix
        if len(args) > 0:
            self.labelStyle = args
        self.label.setHtml(self.labelString())
        self._adjustSize()
        self.picture = None
        self.update()

    def labelString(self):
        if self.labelUnits == "":
            if not self.autoSIPrefix or self.autoSIPrefixScale == 1.0:
                units = ""
            else:
                units = "(x%g)" % (1.0 / self.autoSIPrefixScale)
        else:
            units = "(%s%s)" % (self.labelUnitPrefix, self.labelUnits)

        s = "%s %s" % (self.labelText, units)
        self._label_with_unit = s

        style = ";".join(["%s: %s" % (k, self.labelStyle[k]) for k in self.labelStyle])

        lsbl = "<span style='%s'>%s</span>" % (style, s)
        return lsbl

    def mouseDragEvent(self, event):
        if self.locked:
            return

        if self.orientation in ("left", "right"):
            ev = event
            ev.accept()

            pos = ev.pos()
            lastPos = ev.lastPos()
            dif = pos - lastPos
            dif = dif * -1

            if ev.button() in [
                QtCore.Qt.MouseButton.LeftButton,
                QtCore.Qt.MouseButton.MiddleButton,
            ]:
                if self.orientation in ("left", "right"):
                    scale = (
                        self.range[1] - self.range[0]
                    ) / self.sceneBoundingRect().height()
                    delta = scale * dif.y()
                else:
                    scale = (
                        self.range[1] - self.range[0]
                    ) / self.sceneBoundingRect().width()
                    delta = scale * dif.x()

                self.setRange(self.range[0] - delta, self.range[1] - delta)

            elif ev.button() & QtCore.Qt.MouseButton.RightButton:
                mid = sum(self.range) / 2
                delta = self.range[-1] - self.range[0]

                if self.orientation in ("left", "right"):
                    if dif.y() > 0:
                        delta = 0.94 * delta
                    else:
                        delta = 1.06 * delta

                else:
                    if dif.x() > 0:
                        delta = 0.94 * delta
                    else:
                        delta = 1.06 * delta

                self.setRange(mid - delta / 2, mid + delta / 2)
        else:
            return self.linkedView().mouseDragEvent(event, axis=0, ignore_cursor=True)

    def mouseClickEvent(self, event):
        if event.button() == QtCore.Qt.MouseButton.RightButton:
            event.accept()
            self.raiseContextMenu(event)
        else:
            lv = self.linkedView()
            if lv is None:
                return
            return lv.mouseClickEvent(event)

    def resizeEvent(self, ev=None):
        if self.orientation in ("left", "right"):
            nudge = 5

            if self.minus is not None:
                br = self.minus.boundingRect()
                p = QtCore.QPointF(0, 0)
                if self.orientation == "left":
                    p.setY(5)
                    p.setX(nudge)
                elif self.orientation == "right":
                    p.setY(5)
                    p.setX(int(self.size().width() - br.height() + nudge))

                self.minus.setPos(p)

            if self.plus is not None:
                br = self.plus.boundingRect()
                p = QtCore.QPointF(0, 0)
                if self.orientation == "left":
                    p.setY(26)
                    p.setX(nudge)
                elif self.orientation == "right":
                    p.setY(26)
                    p.setX(int(self.size().width() - br.height() + nudge))

                self.plus.setPos(p)

        return super().resizeEvent(ev)

    def close(self):
        if self.plus is not None:
            self.scene().removeItem(self.plus)
        if self.minus is not None:
            self.scene().removeItem(self.minus)
        self.plus = None
        self.minus = None

        super().close()

    def set_pen(self, pen=None):
        if pen is None:
            pen = fn.mkPen(pen)

        color = pen.color()

        if self.minus is not None:
            p = QtGui.QPainter(self.minus.pixmap)
            self.minus.pixmap.fill(color)
            p.setBrush(self.background)
            p.drawRect(QtCore.QRect(4, 24, 56, 15))
            p.end()

        if self.plus is not None:
            p = QtGui.QPainter(self.plus.pixmap)
            self.plus.pixmap.fill(color)
            p.setBrush(self.background)
            p.drawRect(QtCore.QRect(4, 24, 56, 15))
            p.drawRect(QtCore.QRect(24, 4, 15, 56))
            p.end()

        if pen is not self._pen:
            self.setPen(pen)

    def raiseContextMenu(self, ev):
        low, high = self.range

        if self.orientation in ("left", "right"):
            axis = "Y"
        else:
            axis = "X"

        menu = QtWidgets.QMenu()
        menu.addAction(f"Edit {axis} axis scaling")
        menu.addSeparator()
        menu.addAction("Apply new axis limits")
        menu.addSeparator()

        widget = QtWidgets.QWidget()
        layout = QtWidgets.QHBoxLayout()
        widget.setLayout(layout)
        layout.addWidget(QtWidgets.QLabel("max:"))

        upper = QtWidgets.QDoubleSpinBox()
        upper.setDecimals(9)
        upper.setMinimum(-1e64)
        upper.setMaximum(1e35)
        upper.setValue(high)
        layout.addWidget(upper)

        a = QtWidgets.QWidgetAction(self)
        a.setDefaultWidget(widget)
        menu.addAction(a)

        widget = QtWidgets.QWidget()
        layout = QtWidgets.QHBoxLayout()
        widget.setLayout(layout)
        layout.addWidget(QtWidgets.QLabel("min:"))

        lower = QtWidgets.QDoubleSpinBox()
        lower.setDecimals(9)
        lower.setMinimum(-1e64)
        lower.setMaximum(1e35)
        lower.setValue(low)
        layout.addWidget(lower)

        a = QtWidgets.QWidgetAction(self)
        a.setDefaultWidget(widget)
        menu.addAction(a)

        action = menu.exec_(ev.screenPos().toPoint())

        if action is None:
            return

        elif action.text() == "Apply new axis limits":
            if self.orientation in ("left", "right"):
                self.setRange(lower.value(), upper.value())
            else:
                self.setRange(lower.value(), upper.value())

        elif action.text() == "Edit Y axis scaling":
            self.scale_editor_requested.emit(self.uuid)

    def set_font_size(self, size):
        font = self.font()
        font.setPointSize(size)

        if self.orientation in ("top", "bottom"):
            metric = QtGui.QFontMetrics(font)
            height = max(metric.height() + 2, 18)
            self.setStyle(tickFont=font, tickTextHeight=height)
        else:
            self.setStyle(tickFont=font)
            self.setFont(font)

        self.label.setFont(font)

    def setRange(self, mn, mx):
        if mn > mx:
            mn, mx = mx, mn
        super().setRange(mn, mx)
        self.rangeChanged.emit(self.uuid, (mn, mx))

    def wheelEvent(self, event):
        if self.locked:
            return

        lv = self.linkedView()

        if lv is None:
            # this is one of the individual axis

            pos = event.pos()
            rect = self.boundingRect()

            y_pos_val = ((rect.height() + rect.y()) - pos.y()) / rect.height() * (
                self.range[-1] - self.range[0]
            ) + self.range[0]

            ratio = abs((pos.y() - (rect.height() + rect.y())) / rect.height())

            delta = self.range[-1] - self.range[0]

            if event.delta() > 0:
                delta = 0.66 * delta
            else:
                delta = 1.33 * delta

            start = y_pos_val - ratio * delta
            end = y_pos_val + (1 - ratio) * delta

            self.setRange(start, end)

            event.accept()
        else:
            # this is the main Y axis or the X axis

            if lv.state["mouseMode"] == lv.CursorMode:
                if self.orientation in ("top", "bottom"):
                    super().wheelEvent(event)
                else:
                    # main Y axis
                    if lv.state["mouseEnabled"][1]:
                        # the plot is not Y locked
                        pos = event.pos()
                        rect = self.boundingRect()

                        y_pos_val = (
                            (rect.height() + rect.y()) - pos.y()
                        ) / rect.height() * (
                            self.range[-1] - self.range[0]
                        ) + self.range[
                            0
                        ]

                        ratio = abs(
                            (pos.y() - (rect.height() + rect.y())) / rect.height()
                        )

                        delta = self.range[-1] - self.range[0]

                        if event.delta() > 0:
                            delta = 0.66 * delta
                        else:
                            delta = 1.33 * delta

                        start = y_pos_val - ratio * delta
                        end = y_pos_val + (1 - ratio) * delta

                        self.setRange(start, end)

                        event.accept()
                    else:
                        event.ignore()

            else:
                if self.orientation in ("top", "bottom") or lv.state["mouseEnabled"][1]:
                    super().wheelEvent(event)
                else:
                    event.ignore()

    def paint(self, p, opt, widget):
        rect = self.boundingRect()

        width = rect.width()
        if self.picture is None:
            try:
                picture = QtGui.QPixmap(width, rect.height())
                picture.fill(self.background)

                painter = QtGui.QPainter()
                painter.begin(picture)

                if self.isVisible():
                    painter.setCompositionMode(
                        QtGui.QPainter.CompositionMode_SourceOver
                    )
                    if self.style["tickFont"]:
                        painter.setFont(self.style["tickFont"])
                    specs = self.generateDrawSpecs(painter)

                    if specs is not None:
                        self.drawPicture(painter, *specs)

                    if self.minus is not None:
                        painter.drawPixmap(
                            QtCore.QPoint(int(rect.x()) + 5, 6),
                            self.minus.pixmap.scaled(BUTTON_SIZE, BUTTON_SIZE),
                        )
                        painter.drawPixmap(
                            QtCore.QPoint(int(rect.x()) + 5, 27),
                            self.plus.pixmap.scaled(BUTTON_SIZE, BUTTON_SIZE),
                        )

                    if self.orientation in ("left", "right"):
                        painter.setPen(self._pen)

                        label_rect = QtCore.QRectF(
                            1, 1, rect.height() - (28 + BUTTON_SIZE), rect.width()
                        )
                        painter.translate(rect.bottomLeft())
                        painter.rotate(-90)

                        painter.setRenderHint(painter.RenderHint.TextAntialiasing, True)
                        painter.drawText(
                            label_rect,
                            QtCore.Qt.AlignHCenter | QtCore.Qt.AlignTop,
                            self._label_with_unit,
                        )
                        painter.rotate(90)
                        painter.resetTransform()

            finally:
                painter.end()
            self.picture = picture

    def generateDrawSpecs(self, p):
        """
        Calls tickValues() and tickStrings() to determine where and how ticks should
        be drawn, then generates from this a set of drawing commands to be
        interpreted by drawPicture().
        """
        bounds = self.mapRectFromParent(self.geometry())

        tickBounds = bounds

        if self.orientation == "left":
            span = (bounds.topRight(), bounds.bottomRight())
            tickStart = tickBounds.right()
            tickStop = bounds.right()
            tickDir = -1
            axis = 0
        elif self.orientation == "right":
            span = (bounds.topLeft(), bounds.bottomLeft())
            tickStart = tickBounds.left()
            tickStop = bounds.left()
            tickDir = 1
            axis = 0
        elif self.orientation == "top":
            span = (bounds.bottomLeft(), bounds.bottomRight())
            tickStart = tickBounds.bottom()
            tickStop = bounds.bottom()
            tickDir = -1
            axis = 1
        elif self.orientation == "bottom":
            span = (bounds.topLeft(), bounds.topRight())
            tickStart = tickBounds.top()
            tickStop = bounds.top()
            tickDir = 1
            axis = 1
        else:
            raise ValueError(
                "self.orientation must be in ('left', 'right', 'top', 'bottom')"
            )

        ## determine size of this item in pixels
        points = list(map(self.mapToDevice, span))
        if None in points:
            return
        lengthInPixels = Point(points[1] - points[0]).length()
        if lengthInPixels == 0:
            return

        # Determine major / minor / subminor axis ticks
        if self._tickLevels is None:
            tickLevels = self.tickValues(self.range[0], self.range[1], lengthInPixels)
            tickStrings = None
        else:
            ## parse self.tickLevels into the formats returned by tickLevels() and tickStrings()
            tickLevels = []
            tickStrings = []
            for level in self._tickLevels:
                values = []
                strings = []
                tickLevels.append((None, values))
                tickStrings.append(strings)
                for val, strn in level:
                    values.append(val)
                    strings.append(strn)

        ## determine mapping between tick values and local coordinates
        dif = self.range[1] - self.range[0]
        if dif == 0:
            xScale = 1
            offset = 0
        else:
            if axis == 0:
                xScale = -bounds.height() / dif
                offset = self.range[0] * xScale - bounds.height()
            else:
                xScale = bounds.width() / dif
                offset = self.range[0] * xScale

        xRange = [x * xScale - offset for x in self.range]
        xMin = min(xRange)
        xMax = max(xRange)

        tickPositions = []  # remembers positions of previously drawn ticks

        ## compute coordinates to draw ticks
        ## draw three different intervals, long ticks first
        tickSpecs = []
        for i in range(len(tickLevels)):
            tickPositions.append([])
            ticks = tickLevels[i][1]

            ## length of tick
            tickLength = self.style["tickLength"] / ((i * 0.5) + 1.0)

            lineAlpha = self.style["tickAlpha"]
            if lineAlpha is None:
                lineAlpha = 255 / (i + 1)
                if self.grid is not False:
                    lineAlpha *= (
                        self.grid
                        / 255.0
                        * fn.clip_scalar(
                            (0.05 * lengthInPixels / (len(ticks) + 1)), 0.0, 1.0
                        )
                    )
            elif isinstance(lineAlpha, float):
                lineAlpha *= 255
                lineAlpha = max(0, int(round(lineAlpha)))
                lineAlpha = min(255, int(round(lineAlpha)))
            elif isinstance(lineAlpha, int):
                if (lineAlpha > 255) or (lineAlpha < 0):
                    raise ValueError("lineAlpha should be [0..255]")
            else:
                raise TypeError("Line Alpha should be of type None, float or int")

            for v in ticks:
                ## determine actual position to draw this tick
                x = (v * xScale) - offset
                if (
                    x < xMin or x > xMax
                ):  ## last check to make sure no out-of-bounds ticks are drawn
                    tickPositions[i].append(None)
                    continue
                tickPositions[i].append(x)

                p1 = [x, x]
                p2 = [x, x]
                p1[axis] = tickStart
                p2[axis] = tickStop
                p2[axis] += tickLength * tickDir
                tickPen = self.pen()
                color = tickPen.color()
                color.setAlpha(int(lineAlpha))
                tickPen.setColor(color)
                tickSpecs.append((tickPen, Point(p1), Point(p2)))

        if self.style["stopAxisAtTick"][0] is True:
            minTickPosition = min(map(min, tickPositions))
            if axis == 0:
                stop = max(span[0].y(), minTickPosition)
                span[0].setY(stop)
            else:
                stop = max(span[0].x(), minTickPosition)
                span[0].setX(stop)
        if self.style["stopAxisAtTick"][1] is True:
            maxTickPosition = max(map(max, tickPositions))
            if axis == 0:
                stop = min(span[1].y(), maxTickPosition)
                span[1].setY(stop)
            else:
                stop = min(span[1].x(), maxTickPosition)
                span[1].setX(stop)
        axisSpec = (self.pen(), span[0], span[1])

        textOffset = self.style["tickTextOffset"][
            axis
        ]  ## spacing between axis and text
        # if self.style['autoExpandTextSpace'] is True:
        # textWidth = self.textWidth
        # textHeight = self.textHeight
        # else:
        # textWidth = self.style['tickTextWidth'] ## space allocated for horizontal text
        # textHeight = self.style['tickTextHeight'] ## space allocated for horizontal text

        textSize2 = 0
        lastTextSize2 = 0
        textRects = []
        textSpecs = []  ## list of draw

        # If values are hidden, return early
        if not self.style["showValues"]:
            return (axisSpec, tickSpecs, textSpecs)

        for i in range(min(len(tickLevels), self.style["maxTextLevel"] + 1)):
            ## Get the list of strings to display for this level
            if tickStrings is None:
                spacing, values = tickLevels[i]
                strings = self.tickStrings(
                    values, self.autoSIPrefixScale * self.scale, spacing
                )
            else:
                strings = tickStrings[i]

            if len(strings) == 0:
                continue

            ## ignore strings belonging to ticks that were previously ignored
            for j in range(len(strings)):
                if tickPositions[i][j] is None:
                    strings[j] = None

            ## Measure density of text; decide whether to draw this level
            rects = []
            for s in strings:
                if s is None:
                    rects.append(None)
                else:
                    br = p.boundingRect(
                        QtCore.QRectF(0, 0, 100, 100),
                        QtCore.Qt.AlignmentFlag.AlignCenter,
                        s,
                    )
                    ## boundingRect is usually just a bit too large
                    ## (but this probably depends on per-font metrics?)
                    br.setHeight(br.height() * 0.8)

                    rects.append(br)
                    textRects.append(rects[-1])

            if len(textRects) > 0:
                ## measure all text, make sure there's enough room
                if axis == 0:
                    textSize = np.sum([r.height() for r in textRects])
                    textSize2 = np.max([r.width() for r in textRects])
                else:
                    textSize = np.sum([r.width() for r in textRects])
                    textSize2 = np.max([r.height() for r in textRects])
            else:
                textSize = 0
                textSize2 = 0

            if i > 0:  ## always draw top level
                ## If the strings are too crowded, stop drawing text now.
                ## We use three different crowding limits based on the number
                ## of texts drawn so far.
                textFillRatio = float(textSize) / lengthInPixels
                finished = False
                for nTexts, limit in self.style["textFillLimits"]:
                    if len(textSpecs) >= nTexts and textFillRatio >= limit:
                        finished = True
                        break
                if finished:
                    break

            lastTextSize2 = textSize2

            # spacing, values = tickLevels[best]
            # strings = self.tickStrings(values, self.scale, spacing)
            # Determine exactly where tick text should be drawn
            for j in range(len(strings)):
                vstr = strings[j]
                if vstr is None:  ## this tick was ignored because it is out of bounds
                    continue
                x = tickPositions[i][j]
                # textRect = p.boundingRect(QtCore.QRectF(0, 0, 100, 100), QtCore.Qt.AlignmentFlag.AlignCenter, vstr)
                textRect = rects[j]
                height = textRect.height()
                width = textRect.width()
                # self.textHeight = height
                offset = max(0, self.style["tickLength"]) + textOffset

                if self.orientation == "left":
                    alignFlags = (
                        QtCore.Qt.AlignmentFlag.AlignRight
                        | QtCore.Qt.AlignmentFlag.AlignVCenter
                    )
                    rect = QtCore.QRectF(
                        tickStop - offset - width, x - (height / 2), width, height
                    )
                elif self.orientation == "right":
                    alignFlags = (
                        QtCore.Qt.AlignmentFlag.AlignLeft
                        | QtCore.Qt.AlignmentFlag.AlignVCenter
                    )
                    rect = QtCore.QRectF(
                        tickStop + offset, x - (height / 2), width, height
                    )
                elif self.orientation == "top":
                    alignFlags = (
                        QtCore.Qt.AlignmentFlag.AlignHCenter
                        | QtCore.Qt.AlignmentFlag.AlignBottom
                    )
                    rect = QtCore.QRectF(
                        x - width / 2.0, tickStop - offset - height, width, height
                    )
                elif self.orientation == "bottom":
                    alignFlags = (
                        QtCore.Qt.AlignmentFlag.AlignHCenter
                        | QtCore.Qt.AlignmentFlag.AlignTop
                    )
                    rect = QtCore.QRectF(
                        x - width / 2.0, tickStop + offset, width, height
                    )

                textFlags = alignFlags | QtCore.Qt.TextFlag.TextDontClip
                # p.setPen(self.pen())
                # p.drawText(rect, textFlags, vstr)
                textSpecs.append((rect, textFlags, vstr))

        ## update max text size if needed.
        self._updateMaxTextSize(lastTextSize2)

        self.tickSpecs = tickSpecs

        return (axisSpec, tickSpecs, textSpecs)



# -*- coding: utf-8 -*-

import numpy as np
from PySide6 import QtCore, QtGui, QtWidgets

from asammdf.blocks import v4_constants as v4c
from asammdf.blocks.conversion_utils import from_dict

from ..ui import resource_rc
from ..ui.define_conversion_dialog import Ui_ConversionDialog
from .messagebox import MessageBox


# from https://stackoverflow.com/a/53936965/11009349
def range_overlapping(x, y):
    if x.start == x.stop or y.start == y.stop:
        return False
    return x.start <= y.stop and y.start <= x.stop


class ConversionEditor(Ui_ConversionDialog, QtWidgets.QDialog):
    def __init__(self, channel_name="", conversion=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setupUi(self)

        self.vtt_default_conversion = None
        self.vrtt_default_conversion = None

        self.vtt_default_mode.currentIndexChanged.connect(self.vtt_mode.setCurrentIndex)
        self.vrtt_default_mode.currentIndexChanged.connect(
            self.vrtt_mode.setCurrentIndex
        )

        self.vtt_default_btn.clicked.connect(self.edit_vtt_default_conversion)
        self.vrtt_default_btn.clicked.connect(self.edit_vrtt_default_conversion)

        for widget in (
            self.a,
            self.b,
            self.p1,
            self.p2,
            self.p3,
            self.p4,
            self.p5,
            self.p6,
        ):
            widget.setMaximum(np.inf)
            widget.setMinimum(-np.inf)

        original_conversion_color = QtGui.QColor("#62b2e2")
        bar = self.tabs.tabBar()

        if conversion is not None:
            self.name.setText(conversion["name"])
            self.unit.setText(conversion["unit"])
            self.comment.setPlainText(conversion["comment"])

            if conversion["conversion_type"] == v4c.CONVERSION_TYPE_LIN:
                self.tabs.setCurrentIndex(0)
                bar.setTabTextColor(0, original_conversion_color)

                self.a.setValue(conversion["a"])
                self.b.setValue(conversion["b"])

            elif conversion["conversion_type"] == v4c.CONVERSION_TYPE_RAT:
                self.tabs.setCurrentIndex(1)
                bar.setTabTextColor(1, original_conversion_color)

                self.p1.setValue(conversion["P1"])
                self.p2.setValue(conversion["P2"])
                self.p3.setValue(conversion["P3"])
                self.p4.setValue(conversion["P4"])
                self.p5.setValue(conversion["P5"])
                self.p6.setValue(conversion["P6"])

            elif conversion["conversion_type"] == v4c.CONVERSION_TYPE_TABX:
                self.tabs.setCurrentIndex(2)
                bar.setTabTextColor(2, original_conversion_color)

                if isinstance(conversion.referenced_blocks["default_addr"], bytes):
                    self.vtt_default_mode.setCurrentIndex(0)

                    self.vtt_default.setText(
                        conversion.referenced_blocks["default_addr"].decode(
                            "utf-8", errors="replace"
                        )
                    )
                else:
                    self.vtt_default_mode.setCurrentIndex(1)

                    self.vtt_default_conversion = conversion.referenced_blocks[
                        "default_addr"
                    ]

                for i in range(conversion.ref_param_nr - 1):
                    if isinstance(conversion.referenced_blocks[f"text_{i}"], bytes):
                        widget = VTTWidget(
                            mode="text",
                            value=conversion[f"val_{i}"],
                            text=conversion.referenced_blocks[f"text_{i}"].decode(
                                "utf-8", errors="replace"
                            ),
                        )

                    else:
                        widget = VTTWidget(
                            mode="conversion",
                            value=conversion[f"val_{i}"],
                            conversion=conversion.referenced_blocks[f"text_{i}"],
                        )

                    item = QtWidgets.QListWidgetItem()
                    item.setSizeHint(widget.sizeHint())
                    self.vtt_list.addItem(item)
                    self.vtt_list.setItemWidget(item, widget)

            elif conversion["conversion_type"] == v4c.CONVERSION_TYPE_RTABX:
                self.tabs.setCurrentIndex(3)
                bar.setTabTextColor(3, original_conversion_color)

                if isinstance(conversion.referenced_blocks["default_addr"], bytes):
                    self.vrtt_default_mode.setCurrentIndex(0)

                    self.vrtt_default.setText(
                        conversion.referenced_blocks["default_addr"].decode(
                            "utf-8", errors="replace"
                        )
                    )
                else:
                    self.vrtt_default_mode.setCurrentIndex(1)

                    self.vrtt_default_conversion = conversion.referenced_blocks[
                        "default_addr"
                    ]

                for i in range(conversion.ref_param_nr - 1):
                    if isinstance(conversion.referenced_blocks[f"text_{i}"], bytes):
                        widget = VRTTWidget(
                            mode="text",
                            lower=conversion[f"lower_{i}"],
                            upper=conversion[f"upper_{i}"],
                            text=conversion.referenced_blocks[f"text_{i}"].decode(
                                "utf-8", errors="replace"
                            ),
                        )

                    else:
                        widget = VRTTWidget(
                            mode="conversion",
                            lower=conversion[f"lower_{i}"],
                            upper=conversion[f"upper_{i}"],
                            conversion=conversion.referenced_blocks[f"text_{i}"],
                        )

                    item = QtWidgets.QListWidgetItem()
                    item.setSizeHint(widget.sizeHint())
                    self.vrtt_list.addItem(item)
                    self.vrtt_list.setItemWidget(item, widget)

            elif conversion["conversion_type"] == v4c.CONVERSION_TYPE_NON:
                self.tabs.setCurrentIndex(4)
                bar.setTabTextColor(4, original_conversion_color)
        else:
            self.tabs.setCurrentIndex(4)
            bar.setTabTextColor(4, original_conversion_color)

        self.insert_btn.clicked.connect(self.insert)
        self.insert_vrtt_btn.clicked.connect(self.insert_vrtt)
        self.reset_btn.clicked.connect(self.reset)
        self.reset_vrtt_btn.clicked.connect(self.reset_vrtt)
        self.apply_btn.clicked.connect(self.apply)
        self.cancel_btn.clicked.connect(self.cancel)

        self.vtt_list.setUniformItemSizes(True)
        self.vtt_list.setAlternatingRowColors(False)
        self.vrtt_list.setUniformItemSizes(True)
        self.vrtt_list.setAlternatingRowColors(False)

        self.setWindowTitle(f"Edit {channel_name} conversion")

        self.pressed_button = None

        self.setWindowFlag(QtCore.Qt.WindowMinimizeButtonHint, True)
        self.setWindowFlag(QtCore.Qt.WindowMaximizeButtonHint, True)

    def apply(self, event):
        if self.tabs.currentIndex() == 1:
            if (self.p4.value(), self.p5.value(), self.p6.value()) == (0, 0, 0):
                MessageBox.warning(
                    self,
                    "Invalid conversion parameters",
                    "The rational conversion parameters P4, P5 and P6 cannot all be 0.",
                )
                return

        elif self.tabs.currentIndex() == 2:
            mapping = set()

            for i in range(self.vtt_list.count()):
                item = self.vtt_list.item(i)
                if item is None:
                    continue

                widget = self.vtt_list.itemWidget(item)

                value = int(widget.value.value())

                if value in mapping:
                    MessageBox.warning(
                        self,
                        "Invalid conversion parameters",
                        f"The value-to-text conversion raw value {value} is defined multiple times.",
                    )
                    return
                else:
                    mapping.add(value)

        elif self.tabs.currentIndex() == 3:
            mapping = set()

            for i in range(self.vrtt_list.count()):
                item = self.vrtt_list.item(i)
                if item is None:
                    continue

                widget = self.vrtt_list.itemWidget(item)

                lower = int(widget.lower.value())
                upper = int(widget.upper.value())

                if upper < lower:
                    MessageBox.warning(
                        self,
                        "Invalid conversion parameters",
                        f"The upper value must be greater or higher than the lower value. ({upper=} and {lower=}",
                    )
                    return

                else:
                    y = range(lower, upper)
                    for x in mapping:
                        if range_overlapping(x, y):
                            MessageBox.warning(
                                self,
                                "Invalid conversion parameters",
                                f"The ranges cannot overlap; {x} overlaps with {y}",
                            )
                            return
                        else:
                            mapping.add(y)

        self.pressed_button = "apply"
        self.close()

    def cancel(self, event):
        self.pressed_button = "cancel"
        self.close()

    def conversion(self):
        if self.pressed_button in (None, "cancel"):
            conversion = None

        else:
            if self.tabs.currentIndex() == 0:
                conversion = {
                    "name": self.name.text().strip(),
                    "unit": self.unit.text().strip(),
                    "comment": self.comment.toPlainText().strip(),
                    "a": self.a.value(),
                    "b": self.b.value(),
                }

            elif self.tabs.currentIndex() == 1:
                conversion = {
                    "name": self.name.text().strip(),
                    "unit": self.unit.text().strip(),
                    "comment": self.comment.toPlainText().strip(),
                    "P1": self.p1.value(),
                    "P2": self.p2.value(),
                    "P3": self.p3.value(),
                    "P4": self.p4.value(),
                    "P5": self.p5.value(),
                    "P6": self.p6.value(),
                }

            elif self.tabs.currentIndex() == 2:
                conversion = {
                    "name": self.name.text().strip(),
                    "unit": self.unit.text().strip(),
                    "comment": self.comment.toPlainText().strip(),
                }

                if self.vtt_default_mode.currentIndex() == 0:
                    conversion["default"] = (
                        self.vtt_default.text().strip().encode("utf-8")
                    )
                else:
                    conversion["default"] = self.vtt_default_conversion

                cntr = 0

                for i in range(self.vtt_list.count()):
                    item = self.vtt_list.item(i)
                    if item is None:
                        continue

                    widget = self.vtt_list.itemWidget(item)

                    value = int(widget.value.value())
                    text = widget.reference()
                    conversion[f"val_{cntr}"] = value
                    conversion[f"text_{cntr}"] = text
                    cntr += 1

            elif self.tabs.currentIndex() == 3:
                conversion = {
                    "name": self.name.text().strip(),
                    "unit": self.unit.text().strip(),
                    "comment": self.comment.toPlainText().strip(),
                }

                if self.vrtt_default_mode.currentIndex() == 0:
                    conversion["default"] = (
                        self.vrtt_default.text().strip().encode("utf-8")
                    )
                else:
                    conversion["default"] = self.vrtt_default_conversion

                cntr = 0

                for i in range(self.vrtt_list.count()):
                    item = self.vrtt_list.item(i)
                    if item is None:
                        continue

                    widget = self.vrtt_list.itemWidget(item)

                    lower = int(widget.lower.value())
                    upper = int(widget.upper.value())
                    text = widget.reference()
                    conversion[f"lower_{cntr}"] = lower
                    conversion[f"upper_{cntr}"] = upper
                    conversion[f"text_{cntr}"] = text
                    cntr += 1

            elif self.tabs.currentIndex() == 4:
                conversion = {
                    "name": self.name.text().strip(),
                    "unit": self.unit.text().strip(),
                    "comment": self.comment.toPlainText().strip(),
                }

            conversion = from_dict(conversion)

        return conversion

    def edit_vtt_default_conversion(self):
        dlg = ConversionEditor(f"default", self.vtt_default_conversion, parent=self)
        dlg.exec_()
        if dlg.pressed_button == "apply":
            self.vtt_default_conversion = dlg.conversion()

    def edit_vrtt_default_conversion(self):
        dlg = ConversionEditor(f"default", self.vrtt_default_conversion, parent=self)
        dlg.exec_()
        if dlg.pressed_button == "apply":
            self.vrtt_default_conversion = dlg.conversion()

    def insert(self, event):
        count = self.vtt_list.count()
        if count:
            item = self.vtt_list.item(count - 1)
            last = self.vtt_list.itemWidget(item)
            value = last.value.value() + 1
        else:
            value = 0

        widget = VTTWidget(value=value)

        item = QtWidgets.QListWidgetItem()
        item.setSizeHint(widget.sizeHint())
        self.vtt_list.addItem(item)
        self.vtt_list.setItemWidget(item, widget)

    def insert_vrtt(self, event):
        count = self.vrtt_list.count()
        if count:
            item = self.vrtt_list.item(count - 1)
            last = self.vrtt_list.itemWidget(item)
            lower = last.upper.value()
            upper = lower + 1

        else:
            lower = 0
            upper = 1

        widget = VRTTWidget(lower=lower, upper=upper)

        item = QtWidgets.QListWidgetItem()
        item.setSizeHint(widget.sizeHint())
        self.vrtt_list.addItem(item)
        self.vrtt_list.setItemWidget(item, widget)

    def reset(self, event):
        self.vtt_list.clear()

    def reset_vrtt(self, event):
        self.vrtt_list.clear()


from ..widgets.vrtt_widget import VRTTWidget
from ..widgets.vtt_widget import VTTWidget



# -*- coding: utf-8 -*-
from PySide6 import QtCore, QtWidgets

from ..widgets.bus_database_manager import BusDatabaseManager


class BusDatabaseManagerDialog(QtWidgets.QDialog):
    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        self.setObjectName("BusDatabaseManagerDialog")
        self.resize(404, 294)
        self.setSizeGripEnabled(True)
        self.setWindowFlags(QtCore.Qt.Window)
        self.verticalLayout = QtWidgets.QVBoxLayout(self)

        self._settings = QtCore.QSettings()

        databases = {}

        can_databases = self._settings.value("can_databases", [])
        buses = can_databases[::2]
        dbs = can_databases[1::2]

        databases["CAN"] = [(bus, database) for bus, database in zip(buses, dbs)]

        lin_databases = self._settings.value("lin_databases", [])
        buses = lin_databases[::2]
        dbs = lin_databases[1::2]

        databases["LIN"] = [(bus, database) for bus, database in zip(buses, dbs)]

        self.widget = BusDatabaseManager(databases)

        self.verticalLayout.addWidget(self.widget)

        self.horLayout = QtWidgets.QHBoxLayout(self)

        spacer = QtWidgets.QSpacerItem(
            40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum
        )
        self.apply_btn = QtWidgets.QPushButton("Apply")
        self.cancel_btn = QtWidgets.QPushButton("Cancel")
        self.horLayout.addSpacerItem(spacer)
        self.horLayout.addWidget(self.apply_btn)
        self.horLayout.addWidget(self.cancel_btn)

        self.verticalLayout.addLayout(self.horLayout)

        self.apply_btn.clicked.connect(self.apply)
        self.cancel_btn.clicked.connect(self.cancel)
        self.pressed_button = "cancel"

        self.setWindowTitle("Bus Database Manager")

        self.showMaximized()

    def apply(self, *args):
        self.pressed_button = "apply"
        self.close()

    def cancel(self, *args):
        self.pressed_button = "cancel"
        self.close()

    def store(self):
        databases = self.widget.to_config()

        dbs = []
        for bus, database in databases["CAN"]:
            dbs.append(bus)
            dbs.append(database)

        self._settings.setValue("can_databases", dbs)

        dbs = []
        for bus, database in databases["LIN"]:
            dbs.append(bus)
            dbs.append(database)

        self._settings.setValue("lin_databases", dbs)



import contextlib
from importlib.metadata import distribution, PackageNotFoundError
import re
from typing import Dict, Optional

from packaging.requirements import Requirement
from PySide6.QtCore import QSize
from PySide6.QtGui import QIcon
from PySide6.QtWidgets import QDialog, QTreeWidget, QTreeWidgetItem, QVBoxLayout


class DependenciesDlg(QDialog):
    def __init__(self, package_name: str) -> None:
        """Create a dialog to list all dependencies for `package_name`."""

        super().__init__()

        # Variables
        self._package_name = package_name

        # Widgets
        self._tree = QTreeWidget()

        # Setup widgets
        self._setup_widgets()

        # Setup layout
        self._setup_layout()

        # Signals and slots
        self._connect_signals()

    def _setup_widgets(self) -> None:
        self.setWindowTitle(f"Dependency Overview for {self._package_name}")

        self.setMinimumWidth(700)
        self.setMinimumHeight(500)

        icon = QIcon()
        icon.addFile(":/asammdf.png", QSize(), QIcon.Normal, QIcon.Off)
        self.setWindowIcon(icon)

        headers = ["package", "required", "installed", "summary", "url"]
        self._tree.setHeaderLabels(headers)
        self._populate_tree(self._package_name)
        self._tree.expandAll()
        self._tree.resizeColumnToContents(0)
        self._tree.resizeColumnToContents(1)
        self._tree.resizeColumnToContents(2)

    def _setup_layout(self) -> None:
        vbox = QVBoxLayout()
        self.setLayout(vbox)

        vbox.addWidget(self._tree)

    def _connect_signals(self) -> None:
        self._tree.itemDoubleClicked.connect(self._on_item_double_clicked)

    def _populate_tree(self, package_name: str) -> None:
        package_dist = distribution(package_name)
        requires = package_dist.requires
        if requires is None:
            return

        root_nodes: Dict[str, QTreeWidgetItem] = {}

        def get_root_node(name: Optional[str] = None) -> QTreeWidgetItem:
            if name is None:
                name = "mandatory"

            if name in root_nodes:
                return root_nodes[name]

            new_root_node = QTreeWidgetItem([name])
            root_nodes[name] = new_root_node

            font = new_root_node.font(0)
            font.setBold(True)
            new_root_node.setFont(0, font)

            self._tree.invisibleRootItem().addChild(new_root_node)
            return new_root_node

        for req_string in requires:
            req = Requirement(req_string)

            parent = get_root_node()

            if req.marker is not None:
                match = re.search(
                    r"extra\s*==\s*['\"](?P<extra>\S+)['\"]", str(req.marker)
                )
                if match:
                    parent = get_root_node(match["extra"])

            item = QTreeWidgetItem()
            item.setText(0, req.name)
            item.setText(1, str(req.specifier))

            with contextlib.suppress(PackageNotFoundError):
                dist = distribution(req.name)
                item.setText(2, str(dist.version))
                item.setText(3, dist.metadata["Summary"])
                item.setText(4, dist.metadata["Home-Page"])

            parent.addChild(item)

    def _on_item_double_clicked(self, item: QTreeWidgetItem, column: int) -> None:
        if column != 0:
            return
        if item.parent() is self._tree.invisibleRootItem():
            return

        package_name = item.text(0)
        DependenciesDlg.show_dependencies(package_name)

    @staticmethod
    def show_dependencies(package_name: str) -> None:
        dlg = DependenciesDlg(package_name)
        dlg.exec_()



#!/usr/bin/env python
from pathlib import Path
import tempfile
import unittest

import numpy as np

from asammdf import MDF, Signal
from asammdf.blocks.mdf_v4 import MDF4

CHANNEL_LEN = 100000


class TestMDF4(unittest.TestCase):
    tempdir = None

    @classmethod
    def setUpClass(cls):
        cls.tempdir = tempfile.TemporaryDirectory()

    def test_measurement(self):
        self.assertTrue(MDF4)

    def test_read_mdf4_00(self):
        seed = np.random.randint(0, 2**31)

        np.random.seed(seed)
        print("Read 4.00 using seed =", seed)

        sig_int = Signal(
            np.random.randint(-(2**31), 2**31, CHANNEL_LEN),
            np.arange(CHANNEL_LEN),
            name="Integer Channel",
            unit="unit1",
        )

        sig_float = Signal(
            np.random.random(CHANNEL_LEN),
            np.arange(CHANNEL_LEN),
            name="Float Channel",
            unit="unit2",
        )

        with MDF(version="4.00") as mdf:
            mdf.append([sig_int, sig_float], common_timebase=True)
            outfile = mdf.save(Path(TestMDF4.tempdir.name) / "tmp", overwrite=True)

        with MDF(outfile) as mdf:
            ret_sig_int = mdf.get(sig_int.name)
            ret_sig_float = mdf.get(sig_float.name)

        self.assertTrue(np.array_equal(ret_sig_int.samples, sig_int.samples))
        self.assertTrue(np.array_equal(ret_sig_float.samples, sig_float.samples))

    def test_read_mdf4_10(self):
        seed = np.random.randint(0, 2**31)

        np.random.seed(seed)
        print("Read 4.10 using seed =", seed)

        sig_int = Signal(
            np.random.randint(-(2**31), 2**31, CHANNEL_LEN),
            np.arange(CHANNEL_LEN),
            name="Integer Channel",
            unit="unit1",
        )

        sig_float = Signal(
            np.random.random(CHANNEL_LEN),
            np.arange(CHANNEL_LEN),
            name="Float Channel",
            unit="unit2",
        )

        with MDF(version="4.10") as mdf:
            mdf.append([sig_int, sig_float], common_timebase=True)
            outfile = mdf.save(Path(TestMDF4.tempdir.name) / "tmp", overwrite=True)

        with MDF(outfile) as mdf:
            ret_sig_int = mdf.get(sig_int.name)
            ret_sig_float = mdf.get(sig_float.name)

        self.assertTrue(np.array_equal(ret_sig_int.samples, sig_int.samples))
        self.assertTrue(np.array_equal(ret_sig_float.samples, sig_float.samples))

    def test_attachment_blocks_wo_filename(self):
        original_data = b"Testing attachemnt block\nTest line 1"
        mdf = MDF()
        mdf.attach(
            original_data,
            file_name=None,
            comment=None,
            compression=True,
            mime=r"text/plain",
            embedded=True,
        )
        outfile = mdf.save(
            Path(TestMDF4.tempdir.name) / "attachment.mf4", overwrite=True
        )

        with MDF(outfile) as attachment_mdf:
            data, filename, md5_sum = attachment_mdf.extract_attachment(index=0)
            self.assertEqual(data, original_data)
            self.assertEqual(filename, Path("bin.bin"))

        mdf.close()

    def test_attachment_blocks_w_filename(self):
        original_data = b"Testing attachemnt block\nTest line 1"
        original_file_name = "file.txt"

        mdf = MDF()
        mdf.attach(
            original_data,
            file_name=original_file_name,
            comment=None,
            compression=True,
            mime=r"text/plain",
            embedded=True,
        )
        outfile = mdf.save(
            Path(TestMDF4.tempdir.name) / "attachment.mf4", overwrite=True
        )

        with MDF(outfile) as attachment_mdf:
            data, filename, md5_sum = attachment_mdf.extract_attachment(index=0)
            self.assertEqual(data, original_data)
            self.assertEqual(filename, Path(original_file_name))

        mdf.close()


if __name__ == "__main__":
    unittest.main()



# taken from https://stackoverflow.com/a/46096319/11009349

from PySide6 import QtGui
from PySide6.QtWidgets import QMainWindow, QTreeWidget, QTreeWidgetItem


class ViewTree(QTreeWidget):
    def __init__(self, value, parent=None):
        super().__init__(parent)
        self.setHeaderLabel("Computation parameters")

        def fill_item(item, value):
            def new_item(parent, text, val=None):
                child = QTreeWidgetItem([text])
                fill_item(child, val)
                parent.addChild(child)
                child.setExpanded(True)

            if value is None:
                return
            elif isinstance(value, dict):
                for key, val in sorted(value.items()):
                    new_item(item, str(key), val)
            elif isinstance(value, (list, tuple)):
                for val in value:
                    text = (
                        str(val)
                        if not isinstance(val, (dict, list, tuple))
                        else "[%s]" % type(val).__name__
                    )
                    new_item(item, text, val)
            else:
                new_item(item, str(value))

        fill_item(self.invisibleRootItem(), value)


class ComputedChannelInfoWindow(QMainWindow):
    def __init__(self, signal, parent=None):
        super().__init__(parent)
        self.setCentralWidget(ViewTree(signal.computation, self))
        self.setWindowTitle(f"Computed channel {signal.name}")
        self.setMinimumSize(600, 400)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/info.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)

        self.setWindowIcon(icon)



# -*- coding: utf-8 -*-

from PySide6 import QtWidgets

from ..ui import resource_rc
from ..ui.windows_selection_dialog import Ui_WindowSelectionDialog


class WindowSelectionDialog(Ui_WindowSelectionDialog, QtWidgets.QDialog):
    def __init__(self, options=("Plot", "Numeric", "Tabular"), *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setupUi(self)

        for i, name in enumerate(options):
            radio = QtWidgets.QRadioButton(name)
            self.selection_layout.addWidget(radio)
            if i == 0:
                radio.setChecked(True)

    def selected_type(self):
        for i in range(self.selection_layout.count()):
            radio = self.selection_layout.itemAt(i).widget()
            if radio.isChecked():
                return radio.text()
        return ""

    def disable_new_channels(self):
        return self.disable_channels.isChecked()



# -*- coding: utf-8 -*-
#
# this file contains substantial amount of code from https://github.com/adamerose/PandasGUI which is licensed as MIT:
#
# MIT License
#
# Copyright (c) 2018 Adam Rose
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
#

import bisect
import datetime
import logging
import threading
from traceback import format_exc

import numpy as np
import numpy.core.defchararray as npchar
import pandas as pd
import pyqtgraph.functions as fn
from PySide6 import QtCore, QtGui, QtWidgets

Qt = QtCore.Qt

from ...blocks.utils import (
    csv_bytearray2hex,
    extract_mime_names,
    pandas_query_compatible,
)
from ...mdf import MDF
from ..dialogs.range_editor import RangeEditor
from ..ui import resource_rc
from ..ui.tabular import Ui_TabularDisplay
from ..utils import (
    copy_ranges,
    FONT_SIZE,
    get_colors_using_ranges,
    run_thread_with_progress,
    value_as_str,
)
from .tabular_filter import TabularFilter

logger = logging.getLogger("asammdf.gui")
LOCAL_TIMEZONE = datetime.datetime.now(datetime.timezone.utc).astimezone().tzinfo


MONOSPACE_FONT = None


class TabularTreeItem(QtWidgets.QTreeWidgetItem):
    def __init__(self, column_types, int_format, ranges=None, *args, **kwargs):
        self.column_types = column_types
        self.int_format = int_format
        self.ranges = ranges
        super().__init__(*args, **kwargs)

        self._back_ground_color = self.background(0)
        self._font_color = self.foreground(0)

        self._current_background_color = self._back_ground_color
        self._current_font_color = self._font_color

        self.check_signal_range()

    def __lt__(self, other):
        column = self.treeWidget().sortColumn()

        dtype = self.column_types[column]

        if dtype in "ui":
            if self.int_format == "hex":
                return int(self.text(column), 16) < int(other.text(column), 16)
            elif self.int_format == "bin":
                return int(self.text(column), 2) < int(other.text(column), 2)
            else:
                return int(self.text(column)) < int(other.text(column))

        elif dtype == "f":
            return float(self.text(column)) < float(other.text(column))

        else:
            return self.text(column) < other.text(column)

    def check_signal_range(self):
        if not self.ranges:
            return

        for column, channel_ranges in enumerate(self.ranges, 1):
            value = self.text(column)
            dtype = self.column_types[column]

            if dtype in "ui":
                if self.int_format == "hex":
                    value = int(value, 16)
                elif self.int_format == "bin":
                    value = int(value, 2)
                else:
                    value = int(value)
                value = float(value)

            elif dtype == "f":
                value = float(value)

            new_background_color, new_font_color = get_colors_using_ranges(
                value,
                ranges=channel_ranges,
                default_background_color=self._current_background_color,
                default_font_color=self._current_font_color,
            )

            self.setBackground(column, new_background_color)
            self.setForeground(column, new_font_color)


class DataFrameStorage:
    """
    All methods that modify the data should modify self.df_unfiltered, then self.df gets computed from that
    """

    def __init__(self, df, tabular):
        super().__init__()

        self.df = df
        self.df_unfiltered = df
        self.tabular = tabular

        self.sorted_column_name = None
        self.sorted_index_level = None
        self.sort_state = "None"
        self.dataframe_viewer = None
        self.filter_viewer = None

        self.filters = []
        self.filtered_index_map = df.reset_index().index

        self.data_changed()

    @property
    def sorted_column_ix(self):
        try:
            return list(self.df_unfiltered.columns).index(self.sorted_column_name)
        except ValueError:
            return None

    ###################################
    # Changing columns
    def delete_column(self, ix):
        col_name = self.df_unfiltered.columns[ix]
        self.df_unfiltered = self.df_unfiltered.drop(col_name, axis=1)

        # Need to inform the PyQt model too so column widths properly shift
        self.dataframe_viewer._remove_column(ix)

        self.parent.apply_filters()

    def move_column(self, src, dest):
        cols = list(self.df_unfiltered.columns)
        cols.insert(dest, cols.pop(src))
        self.df_unfiltered = self.df_unfiltered.reindex(cols, axis=1)

        self.dataframe_viewer.setUpdatesEnabled(False)
        # Need to inform the PyQt model too so column widths properly shift
        self.dataframe_viewer._move_column(src, dest)
        self.apply_filters()
        self.dataframe_viewer.setUpdatesEnabled(True)

    ###################################
    # Sorting

    def sort_column(self, ix, next_sort_state=None):
        col_name = self.df_unfiltered.columns[ix]

        # Determine next sorting state by current state
        if next_sort_state is None:
            # Clicked an unsorted column
            if ix != self.sorted_column_ix:
                next_sort_state = "Asc"
            # Clicked a sorted column
            elif ix == self.sorted_column_ix and self.sort_state == "Asc":
                next_sort_state = "Desc"
            # Clicked a reverse sorted column - reset to sorted by index
            elif ix == self.sorted_column_ix:
                next_sort_state = "None"

        if next_sort_state == "Asc":
            self.df_unfiltered = self.df_unfiltered.sort_values(
                col_name, ascending=True, kind="mergesort"
            )
            self.sorted_column_name = self.df_unfiltered.columns[ix]
            self.sort_state = "Asc"

        elif next_sort_state == "Desc":
            self.df_unfiltered = self.df_unfiltered.sort_values(
                col_name, ascending=False, kind="mergesort"
            )
            self.sorted_column_name = self.df_unfiltered.columns[ix]
            self.sort_state = "Desc"

        elif next_sort_state == "None":
            self.df_unfiltered = self.df_unfiltered.sort_index(
                ascending=True, kind="mergesort"
            )
            self.sorted_column_name = None
            self.sort_state = "None"

        self.sorted_index_level = None
        self.tabular.apply_filters()

    def sort_index(self, ix: int):
        # Clicked an unsorted index level
        if ix != self.sorted_index_level:
            self.df_unfiltered = self.df_unfiltered.sort_index(
                level=ix, ascending=True, kind="mergesort"
            )
            self.sorted_index_level = ix
            self.sort_state = "Asc"

        # Clicked a sorted index level
        elif ix == self.sorted_index_level and self.sort_state == "Asc":
            self.df_unfiltered = self.df_unfiltered.sort_index(
                level=ix, ascending=False, kind="mergesort"
            )
            self.sorted_index_level = ix
            self.sort_state = "Desc"

        # Clicked a reverse sorted index level - reset to sorted by full index
        elif ix == self.sorted_index_level:
            self.df_unfiltered = self.df_unfiltered.sort_index(
                ascending=True, kind="mergesort"
            )

            self.sorted_index_level = None
            self.sort_state = "None"

        self.sorted_column = None
        self.tabular.apply_filters()

    def change_column_type(self, ix: int, type):
        name = self.df_unfiltered.columns[ix]
        self.df_unfiltered[name] = self.df_unfiltered[name].astype(type)
        self.tabular.apply_filters()

    ###################################
    # Other

    def data_changed(self):
        self.refresh_ui()

    # Refresh PyQt models when the underlying pgdf is changed in anyway that needs to be reflected in the GUI
    def refresh_ui(self):
        self.models = []

        if self.filter_viewer is not None:
            self.models += [self.filter_viewer.list_model]

        for model in self.models:
            model.beginResetModel()
            model.endResetModel()

        if self.dataframe_viewer is not None:
            self.dataframe_viewer.refresh_ui()


class DataTableModel(QtCore.QAbstractTableModel):
    """
    Model for DataTableView to connect for DataFrame data
    """

    def __init__(self, parent, background_color, font_color):
        super().__init__(parent)
        self.dataframe_viewer = parent
        self.pgdf = parent.pgdf
        self.format = "phys"
        self.float_precision = -1
        self.background_color = background_color
        self.font_color = font_color

    def headerData(self, section, orientation, role=None):
        # Headers for DataTableView are hidden. Header data is shown in HeaderView
        pass

    def columnCount(self, parent=None):
        return self.pgdf.df.columns.shape[0]

    def rowCount(self, parent=None):
        return len(self.pgdf.df)

    # Returns the data from the DataFrame
    def data(self, index, role=QtCore.Qt.DisplayRole):
        row = index.row()
        col = index.column()

        cell = self.pgdf.df.iloc[row, col]

        name = self.pgdf.df_unfiltered.columns[col]

        if role == QtCore.Qt.DisplayRole:
            # Need to check type since a cell might contain a list or Series, then .isna returns a Series not a bool
            cell_is_na = pd.isna(cell)

            if type(cell_is_na) == bool and cell_is_na:
                return "●"
            elif isinstance(cell, (bytes, np.bytes_)):
                return cell.decode("utf-8", "replace")
            else:
                return value_as_str(cell, self.format, None, self.float_precision)

        elif role == QtCore.Qt.BackgroundRole:
            channel_ranges = self.pgdf.tabular.ranges[name]

            try:
                value = float(cell)
            except:
                value = str(cell)

            new_background_color, new_font_color = get_colors_using_ranges(
                value,
                ranges=channel_ranges,
                default_background_color=self.background_color,
                default_font_color=self.font_color,
            )

            return (
                new_background_color
                if new_background_color != self.background_color
                else None
            )

        elif role == QtCore.Qt.ForegroundRole:
            channel_ranges = self.pgdf.tabular.ranges[name]

            try:
                value = float(cell)
            except:
                value = str(cell)

            new_background_color, new_font_color = get_colors_using_ranges(
                value,
                ranges=channel_ranges,
                default_background_color=self.background_color,
                default_font_color=self.font_color,
            )

            return new_font_color if new_font_color != self.font_color else None

        elif role == QtCore.Qt.TextAlignmentRole:
            if isinstance(cell, str):
                return int(QtCore.Qt.AlignLeft | QtCore.Qt.AlignVCenter)
            elif isinstance(cell, pd.Timestamp):
                return int(QtCore.Qt.AlignLeft | QtCore.Qt.AlignVCenter)
            else:
                if self.float_precision == -1 and isinstance(
                    cell, (float, np.floating)
                ):
                    return int(QtCore.Qt.AlignLeft | QtCore.Qt.AlignVCenter)
                else:
                    return int(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)

    def flags(self, index):
        return QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsSelectable

    def setData(self, index, value, role=None):
        pass


class DataTableView(QtWidgets.QTableView):
    add_channels_request = QtCore.Signal(list)

    def __init__(self, parent):
        super().__init__(parent)
        self.dataframe_viewer = parent
        self.pgdf = parent.pgdf

        self._backgrund_color = self.palette().color(QtGui.QPalette.Window)
        self._font_color = self.palette().color(QtGui.QPalette.WindowText)

        # Create and set model
        model = DataTableModel(parent, self._backgrund_color, self._font_color)
        self.setModel(model)

        # Hide the headers. The DataFrame headers (index & columns) will be displayed in the DataFrameHeaderViews
        self.horizontalHeader().hide()
        self.verticalHeader().hide()

        # Link selection to headers
        self.selectionModel().selectionChanged.connect(self.on_selectionChanged)

        # Settings
        # self.setWordWrap(True)
        # self.resizeRowsToContents()
        self.setHorizontalScrollMode(QtWidgets.QAbstractItemView.ScrollPerPixel)
        self.setVerticalScrollMode(QtWidgets.QAbstractItemView.ScrollPerPixel)

        font = QtGui.QFont()
        font.fromString(MONOSPACE_FONT)
        self.setFont(font)

        self.setAcceptDrops(False)
        self.setDragDropMode(QtWidgets.QAbstractItemView.NoDragDrop)
        self.setDropIndicatorShown(False)

    def on_selectionChanged(self):
        """
        Runs when cells are selected in the main table. This logic highlights the correct cells in the vertical and
        horizontal headers when a data cell is selected
        """
        columnHeader = self.dataframe_viewer.columnHeader
        indexHeader = self.dataframe_viewer.indexHeader

        # The two blocks below check what columns or rows are selected in the data table and highlights the
        # corresponding ones in the two headers. The if statements check for focus on headers, because if the user
        # clicks a header that will auto-select all cells in that row or column which will trigger this function
        # and cause and infinite loop

        if not columnHeader.hasFocus():
            selection = self.selectionModel().selection()
            columnHeader.selectionModel().select(
                selection,
                QtCore.QItemSelectionModel.Columns
                | QtCore.QItemSelectionModel.ClearAndSelect,
            )

        if not indexHeader.hasFocus():
            selection = self.selectionModel().selection()
            indexHeader.selectionModel().select(
                selection,
                QtCore.QItemSelectionModel.Rows
                | QtCore.QItemSelectionModel.ClearAndSelect,
            )

    def sizeHint(self):
        # Set width and height based on number of columns in model
        # Width
        width = 2 * self.frameWidth()  # Account for border & padding
        # width += self.verticalScrollBar().width()  # Dark theme has scrollbars always shown
        for i in range(self.model().columnCount()):
            width += self.columnWidth(i)

        # Height
        height = 2 * self.frameWidth()  # Account for border & padding
        # height += self.horizontalScrollBar().height()  # Dark theme has scrollbars always shown
        height += 24 * self.model().rowCount()
        # for i in range(self.model().rowCount()):
        #     height += self.rowHeight(i)

        return QtCore.QSize(width, height)

    def dragEnterEvent(self, e):
        e.accept()

    def dropEvent(self, e):
        if e.source() is self:
            return
        else:
            data = e.mimeData()
            if data.hasFormat("application/octet-stream-asammdf"):
                names = extract_mime_names(data)
                self.add_channels_request.emit(names)
            else:
                return

    def keyPressEvent(self, event):
        key = event.key()
        modifiers = event.modifiers()

        if key == QtCore.Qt.Key_R and modifiers == QtCore.Qt.ControlModifier:
            selected_items = set(
                index.column() for index in self.selectedIndexes() if index.isValid()
            )

            if selected_items:
                ranges = []
                for index in selected_items:
                    original_name = self.pgdf.df_unfiltered.columns[index]
                    ranges.update(self.pgdf.tabular.ranges[original_name])

                dlg = RangeEditor(
                    "<selected signals>", "", ranges=ranges, parent=self, brush=True
                )
                dlg.exec_()
                if dlg.pressed_button == "apply":
                    ranges = dlg.result

                    for index in selected_items:
                        original_name = self.pgdf.df_unfiltered.columns[index]
                        self.pgdf.tabular.ranges[original_name] = copy_ranges(ranges)

        else:
            super().keyPressEvent(event)


class HeaderModel(QtCore.QAbstractTableModel):
    def __init__(self, parent, orientation):
        super().__init__(parent)
        self.orientation = orientation
        self.pgdf = parent.pgdf
        self.prefix = ""

    def columnCount(self, parent=None):
        if self.orientation == Qt.Horizontal:
            return self.pgdf.df.columns.shape[0]
        else:  # Vertical
            return self.pgdf.df.index.nlevels

    def rowCount(self, parent=None):
        if self.orientation == Qt.Horizontal:
            return self.pgdf.df.columns.nlevels
        elif self.orientation == Qt.Vertical:
            return self.pgdf.df.index.shape[0]

    def data(self, index, role=QtCore.Qt.DisplayRole):
        row = index.row()
        col = index.column()

        if role == QtCore.Qt.DisplayRole:
            if self.orientation == Qt.Horizontal:
                if isinstance(self.pgdf.df.columns, pd.MultiIndex):
                    val = str(self.pgdf.df.columns[col][row])
                else:
                    val = str(self.pgdf.df.columns[col])

                val = val[len(self.prefix) :] if val.startswith(self.prefix) else val
                return val

            elif self.orientation == Qt.Vertical:
                if isinstance(self.pgdf.df.index, pd.MultiIndex):
                    return str(self.pgdf.df.index[row][col])
                else:
                    return str(self.pgdf.df.index[row])

        elif role == QtCore.Qt.DecorationRole:
            if self.pgdf.sort_state == "Asc":
                icon = QtGui.QIcon(":/sort-ascending.png")
            elif self.pgdf.sort_state == "Desc":
                icon = QtGui.QIcon(":/sort-descending.png")
            else:
                return

            if (
                col == self.pgdf.sorted_column_ix
                and row == self.rowCount() - 1
                and self.orientation == Qt.Horizontal
            ):
                return icon

        elif role == QtCore.Qt.TextAlignmentRole:
            if self.orientation == Qt.Horizontal:
                name = self.pgdf.df_unfiltered.columns[col]
                dtype = self.pgdf.df_unfiltered[name].values.dtype

                float_precision = (
                    self.pgdf.dataframe_viewer.dataView.model().float_precision
                )

                if np.issubdtype(dtype, np.integer):
                    return int(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
                elif float_precision != -1 and np.issubdtype(dtype, np.floating):
                    return int(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
                else:
                    return int(QtCore.Qt.AlignLeft | QtCore.Qt.AlignVCenter)
            else:
                return QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter

    # The headers of this table will show the level names of the MultiIndex
    def headerData(self, section, orientation, role=None):
        # This was moved to HeaderNamesModel
        pass


class HeaderView(QtWidgets.QTableView):
    """
    Displays the DataFrame index or columns depending on orientation
    """

    def __init__(self, parent, orientation):
        super().__init__(parent)
        self.dataframe_viewer = parent
        self.pgdf = parent.pgdf
        self.setProperty(
            "orientation", "horizontal" if orientation == 1 else "vertical"
        )  # Used in stylesheet

        # Setup
        self.orientation = orientation
        self.table = parent.dataView
        self.setModel(HeaderModel(parent, orientation))
        self.padding = 90

        ###############
        # These are used in self.manage_resizing

        # Holds the index of the cell being resized, or None if resize isn't happening
        self.header_cell_being_resized = None
        # Boolean indicating whether the header itself is currently being resized
        self.header_being_resized = False
        ###############

        # Handled by self.eventFilter()
        self.setMouseTracking(True)
        self.viewport().setMouseTracking(True)
        self.viewport().installEventFilter(self)

        # Settings
        self.setIconSize(QtCore.QSize(16, 16))
        self.setSizePolicy(
            QtWidgets.QSizePolicy(
                QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Maximum
            )
        )
        self.setWordWrap(False)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.setHorizontalScrollMode(QtWidgets.QAbstractItemView.ScrollPerPixel)
        self.setVerticalScrollMode(QtWidgets.QAbstractItemView.ScrollPerPixel)
        font = QtGui.QFont()
        font.fromString(MONOSPACE_FONT)
        font.setBold(True)
        self.setFont(font)

        # Link selection to DataTable
        self.selectionModel().selectionChanged.connect(
            lambda x: self.on_selectionChanged()
        )
        # self.set_spans()

        self.horizontalHeader().hide()
        self.verticalHeader().hide()
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)

        # Automatically stretch rows/columns as widget is resized
        if self.orientation == Qt.Vertical:
            self.horizontalHeader().setSectionResizeMode(
                QtWidgets.QHeaderView.ResizeMode.Stretch
            )

        # Set initial size
        self.resize(self.sizeHint())

    def showEvent(self, a0: QtGui.QShowEvent) -> None:
        super(HeaderView, self).showEvent(a0)
        self.initial_size = self.size()

    def mouseDoubleClickEvent(self, event):
        point = event.pos()
        ix = self.indexAt(point)
        col = ix.column()
        if event.button() == QtCore.Qt.LeftButton:
            # When a header is clicked, sort the DataFrame by that column
            if self.orientation == Qt.Horizontal:
                self.pgdf.sort_column(col)
            else:
                self.on_selectionChanged()
        else:
            super().mouseDoubleClickEvent(event)

    def mousePressEvent(self, event):
        point = event.pos()
        ix = self.indexAt(point)
        col = ix.column()
        if (
            event.button() == QtCore.Qt.RightButton
            and self.orientation == Qt.Horizontal
        ):
            self.dataframe_viewer.show_column_menu(col)
        else:
            super().mousePressEvent(event)

    # Header
    def on_selectionChanged(self, force=False):
        """
        Runs when cells are selected in the Header. This selects columns in the data table when the header is clicked,
        and then calls selectAbove()
        """
        # Check focus so we don't get recursive loop, since headers trigger selection of data cells and vice versa
        if self.hasFocus() or force:
            dataView = self.dataframe_viewer.dataView

            # Set selection mode so selecting one row or column at a time adds to selection each time
            if (
                self.orientation == Qt.Horizontal
            ):  # This case is for the horizontal header
                # Get the header's selected columns
                selection = self.selectionModel().selection()

                # Removes the higher levels so that only the lowest level of the header affects the data table selection
                last_row_ix = self.pgdf.df.columns.nlevels - 1
                last_col_ix = self.model().columnCount() - 1
                higher_levels = QtCore.QItemSelection(
                    self.model().index(0, 0),
                    self.model().index(last_row_ix - 1, last_col_ix),
                )
                selection.merge(higher_levels, QtCore.QItemSelectionModel.Deselect)

                # Select the cells in the data view
                dataView.selectionModel().select(
                    selection,
                    QtCore.QItemSelectionModel.Columns
                    | QtCore.QItemSelectionModel.ClearAndSelect,
                )
            if self.orientation == Qt.Vertical:
                selection = self.selectionModel().selection()

                last_row_ix = self.model().rowCount() - 1
                last_col_ix = self.pgdf.df.index.nlevels - 1
                higher_levels = QtCore.QItemSelection(
                    self.model().index(0, 0),
                    self.model().index(last_row_ix, last_col_ix - 1),
                )
                selection.merge(higher_levels, QtCore.QItemSelectionModel.Deselect)

                dataView.selectionModel().select(
                    selection,
                    QtCore.QItemSelectionModel.Rows
                    | QtCore.QItemSelectionModel.ClearAndSelect,
                )

        self.selectAbove()

    # Take the current set of selected cells and make it so that any spanning cell above a selected cell is selected too
    # This should happen after every selection change
    def selectAbove(self):
        # Disabling this to allow selecting specific cells in headers
        return

        if self.orientation == Qt.Horizontal:
            if self.pgdf.df.columns.nlevels == 1:
                return
        else:
            if self.pgdf.df.index.nlevels == 1:
                return

        for ix in self.selectedIndexes():
            if self.orientation == Qt.Horizontal:
                # Loop over the rows above this one
                for row in range(ix.row()):
                    ix2 = self.model().index(row, ix.column())
                    self.setSelection(
                        self.visualRect(ix2), QtCore.QItemSelectionModel.Select
                    )
            else:
                # Loop over the columns left of this one
                for col in range(ix.column()):
                    ix2 = self.model().index(ix.row(), col)
                    self.setSelection(
                        self.visualRect(ix2), QtCore.QItemSelectionModel.Select
                    )

    # This sets spans to group together adjacent cells with the same values
    def set_spans(self):
        df = self.pgdf.df
        self.clearSpans()
        # Find spans for horizontal HeaderView
        if self.orientation == Qt.Horizontal:
            # Find how many levels the MultiIndex has
            if isinstance(df.columns, pd.MultiIndex):
                N = len(df.columns[0])
            else:
                N = 1

            for level in range(N):  # Iterates over the levels
                # Find how many segments the MultiIndex has
                if isinstance(df.columns, pd.MultiIndex):
                    arr = [df.columns[i][level] for i in range(len(df.columns))]
                else:
                    arr = df.columns

                # Holds the starting index of a range of equal values.
                # None means it is not currently in a range of equal values.
                match_start = None

                for col in range(1, len(arr)):  # Iterates over cells in row
                    # Check if cell matches cell to its left
                    if arr[col] == arr[col - 1]:
                        if match_start is None:
                            match_start = col - 1
                        # If this is the last cell, need to end it
                        if col == len(arr) - 1:
                            match_end = col
                            span_size = match_end - match_start + 1
                            self.setSpan(level, match_start, 1, span_size)
                    else:
                        if match_start is not None:
                            match_end = col - 1
                            span_size = match_end - match_start + 1
                            self.setSpan(level, match_start, 1, span_size)
                            match_start = None

        # Find spans for vertical HeaderView
        else:
            # Find how many levels the MultiIndex has
            if isinstance(df.index, pd.MultiIndex):
                N = len(df.index[0])
            else:
                N = 1

            for level in range(N):  # Iterates over the levels
                # Find how many segments the MultiIndex has
                if isinstance(df.index, pd.MultiIndex):
                    arr = [df.index[i][level] for i in range(len(df.index))]
                else:
                    arr = df.index

                # Holds the starting index of a range of equal values.
                # None means it is not currently in a range of equal values.
                match_start = None

                for row in range(1, len(arr)):  # Iterates over cells in column
                    # Check if cell matches cell above
                    if arr[row] == arr[row - 1]:
                        if match_start is None:
                            match_start = row - 1
                        # If this is the last cell, need to end it
                        if row == len(arr) - 1:
                            match_end = row
                            span_size = match_end - match_start + 1
                            self.setSpan(match_start, level, span_size, 1)
                    else:
                        if match_start is not None:
                            match_end = row - 1
                            span_size = match_end - match_start + 1
                            self.setSpan(match_start, level, span_size, 1)
                            match_start = None

    def eventFilter(self, object: QtCore.QObject, event: QtCore.QEvent):
        if event.type() in [
            QtCore.QEvent.MouseButtonPress,
            QtCore.QEvent.MouseButtonRelease,
            QtCore.QEvent.MouseButtonDblClick,
            QtCore.QEvent.MouseMove,
        ]:
            return self.manage_resizing(object, event)

        return False

    # This method handles all the resizing of headers including column width, row height, and header width/height
    def manage_resizing(self, object: QtCore.QObject, event: QtCore.QEvent):
        # This is used for resizing column widths and row heights
        # For the horizontal header, return the column edge the mouse is over
        # For the vertical header, return the row edge the mouse is over
        # mouse_position is the position along the relevant axis, ie. horizontal x position for the top header
        def over_header_cell_edge(mouse_position, margin=3):
            # Return the index of the column this x position is on the right edge of
            if self.orientation == Qt.Horizontal:
                x = mouse_position
                if self.columnAt(x - margin) != self.columnAt(x + margin):
                    if self.columnAt(x + margin) == 0:
                        # We're at the left edge of the first column
                        return None
                    else:
                        return self.columnAt(x - margin)
                else:
                    return None

            # Return the index of the row this y position is on the top edge of
            elif self.orientation == Qt.Vertical:
                y = mouse_position
                if self.rowAt(y - margin) != self.rowAt(y + margin):
                    if self.rowAt(y + margin) == 0:
                        # We're at the top edge of the first row
                        return None
                    else:
                        return self.rowAt(y - margin)
                else:
                    return None

        # This is used for resizing the left header width or the top header height
        # Returns a boolean indicating whether the mouse is over the header edge to allow resizing
        def over_header_edge(mouse_position: QtCore.QPoint(), margin=7) -> bool:
            if self.orientation == Qt.Horizontal:
                return abs(mouse_position - self.height()) < margin
            elif self.orientation == Qt.Vertical:
                return abs(mouse_position - self.width()) < margin

        # mouse_position is the position along the axis of the header. X pos for top header, Y pos for side header
        if self.orientation == Qt.Horizontal:
            mouse_position = event.pos().x()
            orthogonal_mouse_position = event.pos().y()
        else:
            mouse_position = event.pos().y()
            orthogonal_mouse_position = event.pos().x()

        # Set the cursor shape
        if over_header_cell_edge(mouse_position) is not None:
            if self.orientation == Qt.Horizontal:
                self.viewport().setCursor(QtGui.QCursor(Qt.SplitHCursor))
            elif self.orientation == Qt.Vertical:
                self.viewport().setCursor(QtGui.QCursor(Qt.SplitVCursor))

        elif over_header_edge(orthogonal_mouse_position):
            if self.orientation == Qt.Horizontal:
                # Disabling vertical resizing of top header for now
                pass
                # self.viewport().setCursor(QtGui.QCursor(Qt.SplitVCursor))
            elif self.orientation == Qt.Vertical:
                self.viewport().setCursor(QtGui.QCursor(Qt.SplitHCursor))

        else:
            self.viewport().setCursor(QtGui.QCursor(Qt.ArrowCursor))

        # If mouse is on an edge, start the drag resize process
        if event.type() == QtCore.QEvent.MouseButtonPress:
            if over_header_cell_edge(mouse_position) is not None:
                self.header_cell_being_resized = over_header_cell_edge(mouse_position)
                return True
            # Disabling vertical resizing of top header for now
            elif (
                over_header_edge(orthogonal_mouse_position)
                and self.orientation == Qt.Vertical
            ):
                self.header_being_resized = True
                return True
            else:
                self.header_cell_being_resized = None

        # End the drag process
        if event.type() == QtCore.QEvent.MouseButtonRelease:
            self.header_cell_being_resized = None
            self.header_being_resized = False

        # Auto size the column that was double clicked
        if event.type() == QtCore.QEvent.MouseButtonDblClick:
            # Find which column or row edge the mouse was over and auto size it
            if over_header_cell_edge(mouse_position) is not None:
                header_index = over_header_cell_edge(mouse_position)
                if self.orientation == Qt.Horizontal:
                    self.dataframe_viewer.auto_size_column(header_index)
                elif self.orientation == Qt.Vertical:
                    self.dataframe_viewer.auto_size_row(header_index)
                return True

        # Handle drag resizing
        if event.type() == QtCore.QEvent.MouseMove:
            # If this is None, there is no drag resize happening
            if self.header_cell_being_resized is not None:
                size = mouse_position - self.columnViewportPosition(
                    self.header_cell_being_resized
                )
                if size > 10:
                    if self.orientation == Qt.Horizontal:
                        self.setColumnWidth(self.header_cell_being_resized, size)
                        self.dataframe_viewer.dataView.setColumnWidth(
                            self.header_cell_being_resized, size
                        )
                    if self.orientation == Qt.Vertical:
                        self.setRowHeight(self.header_cell_being_resized, size)
                        self.dataframe_viewer.dataView.setRowHeight(
                            self.header_cell_being_resized, size
                        )

                    self.updateGeometry()
                    self.dataframe_viewer.dataView.updateGeometry()
                return True
            elif self.header_being_resized:
                if self.orientation == Qt.Horizontal:
                    size = orthogonal_mouse_position - self.geometry().top()
                    self.setFixedHeight(max(size, self.initial_size.height()))
                if self.orientation == Qt.Vertical:
                    size = orthogonal_mouse_position - self.geometry().left()
                    self.setFixedWidth(max(size, self.initial_size.width()))

                self.updateGeometry()
                self.dataframe_viewer.dataView.updateGeometry()
                return True
        return False

    # Return the size of the header needed to match the corresponding DataTableView
    def sizeHint(self):
        # Columm headers
        if self.orientation == Qt.Horizontal:
            # Width of DataTableView
            width = self.table.sizeHint().width() + self.verticalHeader().width()
            # Height
            # height = 2 * self.frameWidth()  # Account for border & padding
            # for i in range(self.model().rowCount()):
            #     height += self.rowHeight
            height = 24 * self.model().rowCount()

        # Index header
        else:
            # Height of DataTableView
            height = self.table.sizeHint().height() + self.horizontalHeader().height()
            # Width
            width = 2 * self.frameWidth()  # Account for border & padding
            for i in range(self.model().columnCount()):
                width += max(self.columnWidth(i), 100)
        return QtCore.QSize(width, height)

    # This is needed because otherwise when the horizontal header is a single row it will add whitespace to be bigger
    def minimumSizeHint(self):
        if self.orientation == Qt.Horizontal:
            return QtCore.QSize(0, self.sizeHint().height())
        else:
            return QtCore.QSize(self.sizeHint().width(), 0)


class HeaderNamesModel(QtCore.QAbstractTableModel):
    def __init__(self, parent, orientation):
        super().__init__(parent)
        self.orientation = orientation
        self.pgdf = parent.pgdf
        self.prefix = ""

    def columnCount(self, parent=None):
        if self.orientation == Qt.Horizontal:
            return 1
        elif self.orientation == Qt.Vertical:
            return self.pgdf.df.index.nlevels

    def rowCount(self, parent=None):
        if self.orientation == Qt.Horizontal:
            return self.pgdf.df.columns.nlevels
        elif self.orientation == Qt.Vertical:
            return 1

    def data(self, index, role=QtCore.Qt.DisplayRole):
        row = index.row()
        col = index.column()

        if role == QtCore.Qt.DisplayRole:
            if self.orientation == Qt.Horizontal:
                val = self.pgdf.df.columns.names[row]
                if val is None:
                    val = ""

                val = val[len(self.prefix) :] if val.startswith(self.prefix) else val
                return str(val)

            elif self.orientation == Qt.Vertical:
                val = self.pgdf.df.index.names[col]
                if val is None:
                    val = "Index"
                return str(val)

        elif role == QtCore.Qt.DecorationRole:
            if self.pgdf.sort_state == "Asc":
                icon = QtGui.QIcon(":/sort-ascending.png")
            elif self.pgdf.sort_state == "Desc":
                icon = QtGui.QIcon(":/sort-descending.png")
            else:
                return

            if col == self.pgdf.sorted_index_level and self.orientation == Qt.Vertical:
                return icon

        elif role == QtCore.Qt.TextAlignmentRole:
            return int(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)


class HeaderNamesView(QtWidgets.QTableView):
    def __init__(self, parent, orientation):
        super().__init__(parent)
        self.dataframe_viewer = parent
        self.pgdf = parent.pgdf

        self.setProperty(
            "orientation", "horizontal" if orientation == 1 else "vertical"
        )  # Used in stylesheet

        # Setup
        self.orientation = orientation
        self.setModel(HeaderNamesModel(parent, orientation))

        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)

        self.setHorizontalScrollMode(QtWidgets.QAbstractItemView.ScrollPerPixel)
        self.setVerticalScrollMode(QtWidgets.QAbstractItemView.ScrollPerPixel)

        self.horizontalHeader().hide()
        self.verticalHeader().hide()

        self.setSelectionMode(self.SelectionMode.NoSelection)

        # Automatically stretch rows/columns as widget is resized
        if self.orientation == Qt.Horizontal:
            self.verticalHeader().setSectionResizeMode(
                QtWidgets.QHeaderView.ResizeMode.Stretch
            )
        else:
            self.horizontalHeader().setSectionResizeMode(
                QtWidgets.QHeaderView.ResizeMode.Stretch
            )

        font = QtGui.QFont()
        font.fromString(MONOSPACE_FONT)
        font.setBold(True)
        self.setFont(font)
        self.init_size()

    def mouseDoubleClickEvent(self, event):
        point = event.pos()
        ix = self.indexAt(point)
        if event.button() == QtCore.Qt.LeftButton:
            if self.orientation == Qt.Vertical:
                self.pgdf.sort_index(ix.column())
        else:
            super().mouseDoubleClickEvent(event)

    def init_size(self):
        # Match vertical header name widths to vertical header
        if self.orientation == Qt.Vertical:
            for ix in range(self.model().columnCount()):
                self.setColumnWidth(ix, self.columnWidth(ix))

    def sizeHint(self):
        if self.orientation == Qt.Horizontal:
            width = self.columnWidth(0)
            height = 24
        else:  # Vertical
            width = self.dataframe_viewer.indexHeader.sizeHint().width()
            height = 24

        return QtCore.QSize(width, height)

    def minimumSizeHint(self):
        return self.sizeHint()

    def rowHeight(self, row: int) -> int:
        return self.dataframe_viewer.columnHeader.rowHeight(row)

    def columnWidth(self, column: int) -> int:
        if self.orientation == Qt.Horizontal:
            if all(name is None for name in self.pgdf.df.columns.names):
                return 0
            else:
                return super().columnWidth(column)
        else:
            return self.dataframe_viewer.indexHeader.columnWidth(column)


class ColumnMenu(QtWidgets.QMenu):
    def __init__(self, pgdf, column_ix, parent=None):
        super().__init__(parent)

        self.pgdf = pgdf
        self.column_ix = column_ix

        ########################
        # Info
        idx = self.pgdf.dataframe_viewer.columnHeader.model().index(0, column_ix)
        self.name = self.pgdf.dataframe_viewer.columnHeader.model().data(
            idx, role=QtCore.Qt.DisplayRole
        )
        label = QtWidgets.QLabel(self.name)
        font = QtGui.QFont()
        font.setBold(True)
        label.setFont(font)
        self.add_widget(label)

        self.addSeparator()

        ########################
        # Sorting

        self.add_widget(QtWidgets.QLabel("Set sorting"))

        def select_button():
            self.sort_b1.setDown(
                self.pgdf.sort_state == "Asc"
                and self.pgdf.sorted_column_ix == column_ix
            )
            self.sort_b2.setDown(
                self.pgdf.sort_state == "Desc"
                and self.pgdf.sorted_column_ix == column_ix
            )
            self.sort_b3.setDown(
                self.pgdf.sort_state == "None"
                or self.pgdf.sorted_column_ix != column_ix
            )

        self.sort_b1 = QtWidgets.QPushButton("Asc")
        self.sort_b1.clicked.connect(
            lambda: [self.pgdf.sort_column(self.column_ix, "Asc"), select_button()]
        )

        self.sort_b2 = QtWidgets.QPushButton("Desc")
        self.sort_b2.clicked.connect(
            lambda: [self.pgdf.sort_column(self.column_ix, "Desc"), select_button()]
        )

        self.sort_b3 = QtWidgets.QPushButton("None")
        self.sort_b3.clicked.connect(
            lambda: [self.pgdf.sort_column(self.column_ix, "None"), select_button()]
        )

        select_button()

        sort_control = QtWidgets.QWidget()
        sort_control_layout = QtWidgets.QHBoxLayout()
        sort_control_layout.setSpacing(0)
        sort_control_layout.setContentsMargins(0, 0, 0, 0)
        sort_control.setLayout(sort_control_layout)
        [
            sort_control_layout.addWidget(w)
            for w in [self.sort_b1, self.sort_b2, self.sort_b3]
        ]

        self.add_widget(sort_control)

        self.addSeparator()

        button = QtWidgets.QPushButton("Edit ranges")
        button.clicked.connect(self.edit_ranges)
        self.add_widget(button)

        self.addSeparator()

        action = self.addAction("Automatic set columns width")
        action.triggered.connect(self.automatic_columns_width)

        ########################
        # Move
        #
        # col_name = self.pgdf.df.columns[column_ix]
        # self.move_b1 = QtWidgets.QPushButton("<<")
        # self.move_b1.clicked.connect(lambda: [self.pgdf.move_column(column_ix, 0),
        #                                       self.close(), self.pgdf.dataframe_viewer.show_column_menu(col_name)])
        # self.move_b2 = QtWidgets.QPushButton("<")
        # self.move_b2.clicked.connect(lambda: [self.pgdf.move_column(column_ix, column_ix - 1),
        #                                       self.close(), self.pgdf.dataframe_viewer.show_column_menu(col_name)])
        # self.move_b3 = QtWidgets.QPushButton(">")
        # self.move_b3.clicked.connect(lambda: [self.pgdf.move_column(column_ix, column_ix + 1),
        #                                       self.close(), self.pgdf.dataframe_viewer.show_column_menu(col_name)])
        # self.move_b4 = QtWidgets.QPushButton(">>")
        # self.move_b4.clicked.connect(lambda: [self.pgdf.move_column(column_ix, len(df.columns)),
        #                                       self.close(), self.pgdf.dataframe_viewer.show_column_menu(col_name)])
        #
        # move_control = QtWidgets.QWidget()
        # move_control_layout = QtWidgets.QHBoxLayout()
        # move_control_layout.setSpacing(0)
        # move_control_layout.setContentsMargins(0, 0, 0, 0)
        # move_control.setLayout(move_control_layout)
        # [move_control_layout.addWidget(w) for w in [self.move_b1, self.move_b2, self.move_b3, self.move_b4]]
        #
        # self.add_widget(move_control)

        ########################
        # Delete

        # button = QtWidgets.QPushButton("Delete Column")
        # button.clicked.connect(lambda: [self.pgdf.delete_column(column_ix), self.close()])
        # self.add_widget(button)

        # ########################
        # # Parse Date
        #
        # button = QtWidgets.QPushButton("Parse Date")
        # button.clicked.connect(lambda: [self.pgdf.parse_date(column_ix), self.close()])
        # self.add_widget(button)

        # ########################
        # # Data Type
        # col_type = self.pgdf.df_unfiltered.dtypes[column_ix]
        # types = list(dict.fromkeys([str(col_type)] + ['float', 'bool', 'category', 'str']))
        # combo = QtWidgets.QComboBox()
        # combo.addItems(types)
        # combo.setCurrentText(str(col_type))
        # combo.currentTextChanged.connect(lambda x: [self.pgdf.change_column_type(column_ix, x)])
        # self.add_widget(combo)

        # ########################
        # # Coloring
        # self.add_action("Color by None",
        #                 lambda: [setattr(self.pgdf.dataframe_viewer, 'color_mode', None),
        #                          self.pgdf.dataframe_viewer.refresh_ui()]
        #                 )
        #
        # self.add_action("Color by columns",
        #                 lambda: [setattr(self.pgdf.dataframe_viewer, 'color_mode', 'column'),
        #                          self.pgdf.dataframe_viewer.refresh_ui()]
        #                 )
        #
        # self.add_action("Color by rows",
        #                 lambda: [setattr(self.pgdf.dataframe_viewer, 'color_mode', 'row'),
        #                          self.pgdf.dataframe_viewer.refresh_ui()]
        #                 )
        #
        # self.add_action("Color by all",
        #                 lambda: [setattr(self.pgdf.dataframe_viewer, 'color_mode', 'all'),
        #                          self.pgdf.dataframe_viewer.refresh_ui()]
        #                 )

    def edit_ranges(self, *args):
        self.hide()
        self.pgdf.tabular.edit_ranges(self.column_ix, self.name)

        self.close()

    def add_action(self, text, function):
        action = QtGui.QAction(text, self)
        action.triggered.connect(function)
        self.addAction(action)

    def add_widget(self, widget):
        # https://stackoverflow.com/questions/55086498/highlighting-custom-qwidgetaction-on-hover
        widget.setMouseTracking(True)

        custom_action = QtWidgets.QWidgetAction(self)
        widget.setStyleSheet(widget.styleSheet() + "margin: 5px;")
        custom_action.setDefaultWidget(widget)
        self.addAction(custom_action)

    def show_menu(self, point):
        self.move(point)
        self.show()

    def automatic_columns_width(self, *args):
        self.pgdf.dataframe_viewer.auto_size_header()


class TabularBase(Ui_TabularDisplay, QtWidgets.QWidget):
    add_channels_request = QtCore.Signal(list)
    timestamp_changed_signal = QtCore.Signal(object, float)

    def __init__(self, df, ranges=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setupUi(self)

        if not ranges:
            self.ranges = {name: [] for name in df.columns}
        else:
            self.ranges = {}

            for name, ranges_ in ranges.items():
                for range_info in ranges_:
                    range_info["font_color"] = fn.mkBrush(range_info["font_color"])
                    range_info["background_color"] = fn.mkBrush(
                        range_info["background_color"]
                    )
                self.ranges[name] = ranges_

        df = DataFrameStorage(df, self)

        self.tree = DataFrameViewer(df)

        self.tree.dataView.selectionModel().currentChanged.connect(self.current_changed)
        self.tree.dataView.add_channels_request.connect(self.add_channels_request)

        self.horizontalLayout.insertWidget(0, self.tree)

        self.add_filter_btn.clicked.connect(self.add_filter)
        self.apply_filters_btn.clicked.connect(self.apply_filters)
        self.time_as_date.stateChanged.connect(self.time_as_date_changed)
        self.remove_prefix.stateChanged.connect(self.remove_prefix_changed)
        self.format_selection.currentTextChanged.connect(self.set_format)

        self.toggle_filters_btn.clicked.connect(self.toggle_filters)
        self.filters_group.setHidden(True)

        self.float_precision.addItems(
            ["Full float precision"] + [f"{i} float decimals" for i in range(16)]
        )
        self.float_precision.setCurrentIndex(0)
        self.float_precision.currentIndexChanged.connect(self.float_precision_changed)

        self._timestamps = None

        # self.show()

        self.tree.auto_size_header()

    def float_precision_changed(self, index):
        self.tree.dataView.model().float_precision = index - 1
        self.tree.pgdf.data_changed()

    def current_changed(self, current, previous):
        if current.isValid():
            row = current.row()
            self._filtered_ts_series = self._original_ts_series.reindex(
                self.tree.pgdf.df.index
            )
            ts = float(self._filtered_ts_series.iloc[row])
            self.timestamp_changed_signal.emit(self, ts)

    def toggle_filters(self, event=None):
        if self.toggle_filters_btn.text() == "Show filters":
            self.toggle_filters_btn.setText("Hide filters")
            self.filters_group.setHidden(False)
            icon = QtGui.QIcon()
            icon.addPixmap(
                QtGui.QPixmap(":/up.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off
            )
            self.toggle_filters_btn.setIcon(icon)
        else:
            self.toggle_filters_btn.setText("Show filters")
            self.filters_group.setHidden(True)
            icon = QtGui.QIcon()
            icon.addPixmap(
                QtGui.QPixmap(":/down.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off
            )
            self.toggle_filters_btn.setIcon(icon)

    def add_filter(self, event=None):
        filter_widget = TabularFilter(
            [
                (
                    self.tree.pgdf.df_unfiltered.index.name,
                    self.tree.pgdf.df_unfiltered.index.values.dtype.kind,
                    0,
                    False,
                )
            ]
            + [
                (
                    name,
                    self.tree.pgdf.df_unfiltered[name].values.dtype.kind,
                    self.signals_descr[name],
                )
                for name in self.tree.pgdf.df_unfiltered.columns
            ],
            self.format_selection.currentText(),
        )

        item = QtWidgets.QListWidgetItem(self.filters)
        item.setSizeHint(filter_widget.sizeHint())
        self.filters.addItem(item)
        self.filters.setItemWidget(item, filter_widget)

    def apply_filters(self, event=None):
        df = self.tree.pgdf.df_unfiltered.copy()

        friendly_names = {name: pandas_query_compatible(name) for name in df.columns}

        original_names = {val: key for key, val in friendly_names.items()}

        df.rename(columns=friendly_names, inplace=True)

        filters = []
        count = self.filters.count()

        for i in range(count):
            filter = self.filters.itemWidget(self.filters.item(i))
            if filter.enabled.checkState() == QtCore.Qt.Unchecked:
                continue

            target = filter._target
            if target is None:
                continue

            if filters:
                filters.append(filter.relation.currentText().lower())

            column_name = filter.column.currentText()
            if column_name == df.index.name:
                is_byte_array = False
            else:
                is_byte_array = self.signals_descr[column_name]
            column_name = pandas_query_compatible(column_name)
            op = filter.op.currentText()

            if target != target:
                # here we have NaN
                nan = np.nan

                if op in (">", ">=", "<", "<="):
                    filters.append(column_name)
                    filters.append(op)
                    filters.append("@nan")
                elif op == "!=":
                    filters.append(column_name)
                    filters.append("==")
                    filters.append(column_name)
                elif op == "==":
                    filters.append(column_name)
                    filters.append("!=")
                    filters.append(column_name)
            else:
                if column_name == "timestamps" and df["timestamps"].dtype.kind == "M":
                    ts = pd.Timestamp(target, tz=LOCAL_TIMEZONE)
                    ts = ts.tz_convert("UTC").to_datetime64()

                    filters.append(column_name)
                    filters.append(op)
                    filters.append("@ts")

                elif is_byte_array:
                    target = str(target).replace(" ", "").strip('"')

                    if f"{column_name}__as__bytes" not in df.columns:
                        df[f"{column_name}__as__bytes"] = pd.Series(
                            [bytes(s) for s in df[column_name]], index=df.index
                        )
                    val = bytes.fromhex(target)

                    filters.append(f"{column_name}__as__bytes")
                    filters.append(op)
                    filters.append("@val")

                else:
                    filters.append(column_name)
                    filters.append(op)
                    filters.append(str(target))

        if filters:
            try:
                new_df = df.query(" ".join(filters))
            except:
                logger.exception(
                    f'Failed to apply filter for tabular window: {" ".join(filters)}'
                )
                self.query.setText(format_exc())
            else:
                to_drop = [name for name in df.columns if name.endswith("__as__bytes")]
                if to_drop:
                    df.drop(columns=to_drop, inplace=True)
                    new_df.drop(columns=to_drop, inplace=True)
                self.query.setText(" ".join(filters))
                new_df.rename(columns=original_names, inplace=True)
                self.tree.pgdf.df = new_df
                self.tree.pgdf.data_changed()
        else:
            self.query.setText("")
            df.rename(columns=original_names, inplace=True)
            self.tree.pgdf.df = df
            self.tree.pgdf.data_changed()

        self.tree.pgdf.df_unfiltered.rename(columns=original_names, inplace=True)

    def add_new_channels(self, signals, mime_data=None):
        if len(self.tree.pgdf.df_unfiltered) != len(self.tree.pgdf.df):
            filtered = True
        else:
            filtered = False

        index = pd.Series(np.arange(len(signals), dtype="u8"), index=signals.index)
        signals["Index"] = index

        signals.set_index(index, inplace=True)
        dropped = {}

        ranges = {
            name: channel_ranges
            for name, channel_ranges in zip(
                self.tree.pgdf.df_unfiltered.columns, self.ranges.values()
            )
        }

        for name_ in signals.columns:
            col = signals[name_]
            if col.dtype.kind == "O":
                if name_.endswith("DataBytes"):
                    try:
                        sizes = signals[name_.replace("DataBytes", "DataLength")]
                    except:
                        sizes = None
                    dropped[name_] = pd.Series(
                        csv_bytearray2hex(
                            col,
                            sizes,
                        ),
                        index=signals.index,
                    )

                elif name_.endswith("Data Bytes"):
                    try:
                        sizes = signals[name_.replace("Data Bytes", "Data Length")]
                    except:
                        sizes = None
                    dropped[name_] = pd.Series(
                        csv_bytearray2hex(
                            col,
                            sizes,
                        ),
                        index=signals.index,
                    )

                elif col.dtype.name != "category":
                    try:
                        dropped[name_] = pd.Series(
                            csv_bytearray2hex(col), index=signals.index
                        )
                    except:
                        pass

                self.signals_descr[name_] = 0

            elif col.dtype.kind == "S":
                try:
                    dropped[name_] = pd.Series(
                        npchar.decode(col, "utf-8"), index=signals.index
                    )
                except:
                    dropped[name_] = pd.Series(
                        npchar.decode(col, "latin-1"), index=signals.index
                    )
                self.signals_descr[name_] = 0
            else:
                self.signals_descr[name_] = 0

            ranges[name_] = []

        signals = signals.drop(columns=["Index", *list(dropped)])
        for name, s in dropped.items():
            signals[name] = s

        names = list(signals.columns)
        names = [
            *[name for name in names if name.endswith((".ID", ".DataBytes"))],
            *[
                name
                for name in names
                if name != "timestamps" and not name.endswith((".ID", ".DataBytes"))
            ],
        ]
        signals = signals[names]

        self.tree.pgdf.df_unfiltered = self.tree.pgdf.df = pd.concat(
            [self.tree.pgdf.df_unfiltered, signals], axis=1
        )
        self.ranges = ranges

        if filtered:
            self.apply_filters()
        else:
            self.tree.pgdf.data_changed()

        self.tree.auto_size_header()
        self.tree.update_horizontal_scroll()

    def to_config(self):
        count = self.filters.count()

        pattern = self.pattern
        if pattern:
            ranges = copy_ranges(pattern["ranges"])

            for range_info in ranges:
                range_info["font_color"] = range_info["font_color"].color().name()
                range_info["background_color"] = (
                    range_info["background_color"].color().name()
                )

            pattern["ranges"] = ranges

        ranges = {}
        for name, channel_ranges in self.ranges.items():
            channel_ranges = copy_ranges(channel_ranges)

            for range_info in channel_ranges:
                range_info["font_color"] = range_info["font_color"].color().name()
                range_info["background_color"] = (
                    range_info["background_color"].color().name()
                )

            ranges[name] = channel_ranges

        config = {
            "sorted": True,
            "channels": list(self.tree.pgdf.df_unfiltered.columns)
            if not self.pattern
            else [],
            "filtered": bool(self.query.toPlainText()),
            "filters": [
                self.filters.itemWidget(self.filters.item(i)).to_config()
                for i in range(count)
            ]
            if not self.pattern
            else [],
            "time_as_date": self.time_as_date.checkState() == QtCore.Qt.Checked,
            "pattern": pattern,
            "format": self.format,
            "ranges": ranges,
            "header_sections_width": [
                self.tree.columnHeader.horizontalHeader().sectionSize(i)
                for i in range(self.tree.columnHeader.horizontalHeader().count())
            ],
            "font_size": self.tree.dataView.font().pointSize(),
        }

        return config

    def time_as_date_changed(self, state):
        s = self.start
        count = self.filters.count()

        if state == QtCore.Qt.Checked:
            for i in range(count):
                filter = self.filters.itemWidget(self.filters.item(i))
                filter.dtype_kind[0] = "M"

                if filter.column.currentIndex() == 0:
                    filter.column_changed(0)
                else:
                    filter.validate_target()

            delta = pd.to_timedelta(
                self.tree.pgdf.df_unfiltered["timestamps"], unit="s"
            )

            timestamps = self.start + delta
            self.tree.pgdf.df_unfiltered["timestamps"] = timestamps
        else:
            for i in range(count):
                filter = self.filters.itemWidget(self.filters.item(i))
                filter.dtype_kind[0] = "f"

                if filter.column.currentIndex() == 0:
                    filter.column_changed(0)
                else:
                    filter.validate_target()
            self.tree.pgdf.df_unfiltered["timestamps"] = self._original_timestamps

        if self.query.toPlainText():
            self.apply_filters()

        self.tree.pgdf.data_changed()
        self.tree.auto_size_column(0)

    def remove_prefix_changed(self, state):
        if state == QtCore.Qt.Checked:
            self.prefix.setEnabled(True)

            self.tree.columnHeaderNames.model().prefix = self.prefix.currentText()
            self.tree.columnHeader.model().prefix = self.prefix.currentText()
            self.tree.prefix = self.prefix.currentText()

        else:
            self.prefix.setEnabled(False)

            self.tree.columnHeaderNames.model().prefix = ""
            self.tree.columnHeader.model().prefix = ""
            self.tree.prefix = ""

        self.tree.auto_size_header()

    def prefix_changed(self, index):
        self.remove_prefix_changed(QtCore.Qt.Checked)

    def open_menu(self, position):
        menu = QtWidgets.QMenu()

        menu.addAction(self.tr("Export to CSV"))

        action = menu.exec_(self.tree.viewport().mapToGlobal(position))

        if action is None:
            return

        if action.text() == "Export to CSV":
            file_name, _ = QtWidgets.QFileDialog.getSaveFileName(
                self,
                "Select output CSV file",
                "",
                "CSV (*.csv)",
            )

            if file_name:
                self.progress = 0, 0
                progress = QtWidgets.QProgressDialog(
                    f'Data export to CSV file "{file_name}"', "", 0, 0, self.parent()
                )

                progress.setWindowModality(QtCore.Qt.ApplicationModal)
                progress.setCancelButton(None)
                progress.setAutoClose(True)
                progress.setWindowTitle("Export tabular window to CSV")
                icon = QtGui.QIcon()
                icon.addPixmap(
                    QtGui.QPixmap(":/csv.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off
                )
                progress.setWindowIcon(icon)
                progress.show()

                target = self.tree.pgdf.df_unfiltered.to_csv
                kwargs = {
                    "path_or_buf": file_name,
                    "index_label": "Index",
                    "date_format": "%Y-%m-%d %H:%M:%S.%f%z",
                }

                result = run_thread_with_progress(
                    self,
                    target=target,
                    kwargs=kwargs,
                    factor=0,
                    offset=0,
                    progress=progress,
                )

                progress.cancel()

    def keyPressEvent(self, event):
        key = event.key()
        modifiers = event.modifiers()

        if (
            key in (QtCore.Qt.Key_H, QtCore.Qt.Key_B, QtCore.Qt.Key_P)
            and modifiers == QtCore.Qt.ControlModifier
        ):
            if key == QtCore.Qt.Key_H:
                self.format_selection.setCurrentText("hex")
            elif key == QtCore.Qt.Key_B:
                self.format_selection.setCurrentText("bin")
            else:
                self.format_selection.setCurrentText("phys")

            event.accept()

        elif key == QtCore.Qt.Key_S and modifiers == QtCore.Qt.ControlModifier:
            file_name, _ = QtWidgets.QFileDialog.getSaveFileName(
                self,
                "Select output measurement file",
                "",
                "MDF version 4 files (*.mf4)",
            )

            if file_name:
                with MDF() as mdf:
                    mdf.append(self.tree.pgdf.df_unfiltered)
                    mdf.save(file_name, overwrite=True)

        elif (
            key == QtCore.Qt.Key_BracketLeft and modifiers == QtCore.Qt.ControlModifier
        ):
            self.decrease_font()

        elif (
            key == QtCore.Qt.Key_BracketRight and modifiers == QtCore.Qt.ControlModifier
        ):
            self.increase_font()

        elif key == QtCore.Qt.Key_G and modifiers == QtCore.Qt.ShiftModifier:
            value, ok = QtWidgets.QInputDialog.getDouble(
                self,
                "Go to time stamp",
                "Time stamp",
                decimals=9,
            )

            if ok:
                self.set_timestamp(value)

        else:
            self.tree.dataView.keyPressEvent(event)

    def set_format(self, fmt):
        self.format = fmt

        self.tree.dataView.model().format = fmt
        self.tree.pgdf.data_changed()
        self._settings.setValue("tabular_format", fmt)

        for row in range(self.filters.count()):
            filter = self.filters.itemWidget(self.filters.item(row))
            filter.int_format = fmt
            filter.validate_target()

        if self.query.toPlainText():
            self.apply_filters()

    def set_timestamp(self, stamp):
        self._filtered_ts_series = self._original_ts_series.reindex(
            self.tree.pgdf.df.index
        )

        if not len(self._filtered_ts_series):
            return

        if not (
            self._filtered_ts_series.iloc[0]
            <= stamp
            <= self._filtered_ts_series.iloc[-1]
        ):
            return

        idx = self._filtered_ts_series.searchsorted(stamp, side="right") - 1
        if idx < 0:
            idx = 0

        self.tree.dataView.selectRow(idx)

    def edit_ranges(self, index, name):
        if index >= 0:
            original_name = self.tree.pgdf.df_unfiltered.columns[index]

            dlg = RangeEditor(
                name, "", self.ranges[original_name], parent=self, brush=True
            )
            dlg.exec_()
            if dlg.pressed_button == "apply":
                ranges = dlg.result
                self.ranges[original_name] = ranges

    def decrease_font(self):
        font = self.tree.dataView.font()
        size = font.pointSize()
        pos = bisect.bisect_left(FONT_SIZE, size) - 1
        if pos < 0:
            pos = 0
        new_size = FONT_SIZE[pos]

        self.set_font_size(new_size)

    def increase_font(self):
        font = self.tree.dataView.font()
        size = font.pointSize()
        pos = bisect.bisect_right(FONT_SIZE, size)
        if pos == len(FONT_SIZE):
            pos -= 1
        new_size = FONT_SIZE[pos]

        self.set_font_size(new_size)

    def set_font_size(self, size):
        self.hide()
        font = self.tree.dataView.font()
        font.setPointSize(size)
        self.tree.dataView.setFont(font)
        font.setBold(True)
        self.tree.indexHeader.setFont(font)
        self.tree.indexHeaderNames.setFont(font)
        self.tree.columnHeader.setFont(font)
        self.tree.columnHeaderNames.setFont(font)
        self.show()

        self.tree.default_row_height = 12 + size
        self.tree.set_styles()


class DataFrameViewer(QtWidgets.QWidget):
    def __init__(self, pgdf):
        super().__init__()

        global MONOSPACE_FONT

        if MONOSPACE_FONT is None:
            families = QtGui.QFontDatabase().families()
            for family in (
                "Consolas",
                "Liberation Mono",
                "DejaVu Sans Mono",
                "Droid Sans Mono",
                "Liberation Mono",
                "Roboto Mono",
                "Monaco",
                "Courier",
            ):
                if family in families:
                    MONOSPACE_FONT = (
                        f"{family},9,-1,5,400,0,0,0,0,0,0,0,0,0,0,1,Regular"
                    )
                    break

        pgdf.dataframe_viewer = self
        self.pgdf = pgdf
        self.prefix = ""

        # Local state
        # How to color cells
        self.color_mode = None

        # Set up DataFrame TableView and Model
        self.dataView = DataTableView(parent=self)

        # Create headers
        self.columnHeader = HeaderView(parent=self, orientation=Qt.Horizontal)
        self.indexHeader = HeaderView(parent=self, orientation=Qt.Vertical)

        self.columnHeaderNames = HeaderNamesView(parent=self, orientation=Qt.Horizontal)
        self.indexHeaderNames = HeaderNamesView(parent=self, orientation=Qt.Vertical)

        # Set up layout
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setContentsMargins(0, 0, 0, 0)
        self.gridLayout.setSpacing(0)
        self.setLayout(self.gridLayout)

        # Linking scrollbars
        # Scrolling in data table also scrolls the headers
        self.dataView.horizontalScrollBar().valueChanged.connect(
            self.columnHeader.horizontalScrollBar().setValue
        )
        self.dataView.horizontalScrollBar().valueChanged.connect(
            self.columnHeaderNames.horizontalScrollBar().setValue
        )
        self.dataView.verticalScrollBar().valueChanged.connect(
            self.indexHeader.verticalScrollBar().setValue
        )
        # Scrolling in headers also scrolls the data table
        self.columnHeader.horizontalScrollBar().valueChanged.connect(
            self.dataView.horizontalScrollBar().setValue
        )
        self.indexHeader.verticalScrollBar().valueChanged.connect(
            self.dataView.verticalScrollBar().setValue
        )
        # Turn off default scrollbars
        self.dataView.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.dataView.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        # Disable scrolling on the headers. Even though the scrollbars are hidden, scrolling by dragging desyncs them
        self.indexHeader.horizontalScrollBar().valueChanged.connect(lambda: None)

        class CornerWidget(QtWidgets.QWidget):
            def __init__(self):
                super().__init__()
                # https://stackoverflow.com/questions/32313469/stylesheet-in-pyside-not-working
                self.setAttribute(QtCore.Qt.WA_StyledBackground)

        self.corner_widget = CornerWidget()
        self.corner_widget.setSizePolicy(
            QtWidgets.QSizePolicy(
                QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding
            )
        )
        # Add items to grid layout
        # self.gridLayout.addWidget(self.corner_widget, 0, 0)
        self.gridLayout.addWidget(self.columnHeader, 0, 1, 2, 2, Qt.AlignTop)
        self.gridLayout.addWidget(self.columnHeaderNames, 0, 3, 2, 1)
        self.gridLayout.addWidget(self.indexHeader, 2, 0, 2, 2, Qt.AlignLeft)
        self.gridLayout.addWidget(self.indexHeaderNames, 1, 0, 1, 1, Qt.AlignBottom)
        self.gridLayout.addWidget(self.dataView, 3, 2, 1, 1)
        self.gridLayout.addWidget(self.dataView.horizontalScrollBar(), 4, 2, 1, 1)
        self.gridLayout.addWidget(self.dataView.verticalScrollBar(), 3, 3, 1, 1)

        # Fix scrollbars forcing a minimum height of the dataView which breaks layout for small number of rows
        self.dataView.verticalScrollBar().setSizePolicy(
            QtWidgets.QSizePolicy(
                QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Ignored
            )
        )
        self.dataView.horizontalScrollBar().setSizePolicy(
            QtWidgets.QSizePolicy(
                QtWidgets.QSizePolicy.Ignored, QtWidgets.QSizePolicy.Fixed
            )
        )

        # These expand when the window is enlarged instead of having the grid squares spread out
        # self.gridLayout.setColumnStretch(4, 1)
        # self.gridLayout.setRowStretch(5, 1)

        self.gridLayout.setColumnStretch(2, 1)
        self.gridLayout.setRowStretch(5, 1)

        #
        # self.gridLayout.addItem(QtWidgets.QSpacerItem(0, 0,
        #                                               QtWidgets.QSizePolicy.Expanding,
        #                                               QtWidgets.QSizePolicy.Expanding), 0, 0, 1, 1, )

        self.default_row_height = 24
        self.set_styles()
        self.indexHeader.setSizePolicy(
            QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Maximum
        )
        self.columnHeader.setSizePolicy(
            QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.MinimumExpanding
        )

        # Set column widths
        for column_index in range(self.columnHeader.model().columnCount()):
            self.auto_size_column(column_index)

        self.columnHeader.horizontalHeader().setStretchLastSection(True)
        self.columnHeaderNames.horizontalHeader().setStretchLastSection(True)

        self.columnHeader.horizontalHeader().sectionResized.connect(
            self.update_horizontal_scroll
        )

        self.columnHeader.horizontalHeader().setMinimumSectionSize(1)
        self.dataView.horizontalHeader().setMinimumSectionSize(1)

        # self.show()

        self.auto_size_header()

    def set_styles(self):
        for item in [
            self.dataView,
            self.columnHeader,
            self.indexHeader,
            self.indexHeaderNames,
            self.columnHeaderNames,
        ]:
            item.setContentsMargins(0, 0, 0, 0)
            # item.setItemDelegate(NoFocusDelegate())

        self.indexHeaderNames.verticalHeader().setDefaultSectionSize(
            self.default_row_height
        )
        self.indexHeaderNames.verticalHeader().setMinimumSectionSize(
            self.default_row_height
        )
        self.indexHeaderNames.verticalHeader().setMaximumSectionSize(
            self.default_row_height
        )
        self.indexHeaderNames.verticalHeader().sectionResizeMode(
            QtWidgets.QHeaderView.Fixed
        )
        self.indexHeader.verticalHeader().setDefaultSectionSize(self.default_row_height)
        self.indexHeader.verticalHeader().setMinimumSectionSize(self.default_row_height)
        self.indexHeader.verticalHeader().setMaximumSectionSize(self.default_row_height)
        self.indexHeader.verticalHeader().sectionResizeMode(QtWidgets.QHeaderView.Fixed)
        self.dataView.verticalHeader().setDefaultSectionSize(self.default_row_height)
        self.dataView.verticalHeader().setMinimumSectionSize(self.default_row_height)
        self.dataView.verticalHeader().setMaximumSectionSize(self.default_row_height)
        self.dataView.verticalHeader().sectionResizeMode(QtWidgets.QHeaderView.Fixed)
        self.columnHeader.verticalHeader().setDefaultSectionSize(
            self.default_row_height
        )
        self.columnHeader.verticalHeader().setMinimumSectionSize(
            self.default_row_height
        )
        self.columnHeader.verticalHeader().setMaximumSectionSize(
            self.default_row_height
        )
        self.columnHeader.verticalHeader().sectionResizeMode(
            QtWidgets.QHeaderView.Fixed
        )

    def __reduce__(self):
        # This is so dataclasses.asdict doesn't complain about this being unpicklable
        return "DataFrameViewer"

    def auto_size_header(self):
        s = 0
        for i in range(self.columnHeader.model().columnCount()):
            s += self.auto_size_column(i)

        delta = int((self.dataView.size().width() - s) // len(self.pgdf.df.columns))

        if delta > 0:
            for i in range(self.columnHeader.model().columnCount()):
                self.auto_size_column(i, extra_padding=delta)
            self.dataView.horizontalScrollBar().hide()
        else:
            self.dataView.horizontalScrollBar().show()

    def update_horizontal_scroll(self, *args):
        s = 0
        for i in range(self.columnHeader.model().columnCount()):
            s += self.columnHeader.columnWidth(i) + self.columnHeader.frameWidth()

        if self.dataView.size().width() < s:
            self.dataView.horizontalScrollBar().show()
        else:
            self.dataView.horizontalScrollBar().hide()

    def auto_size_column(self, column_index, extra_padding=0):
        """
        Set the size of column at column_index to fit its contents
        """

        width = 0

        # Iterate over the data view rows and check the width of each to determine the max width for the column
        # Only check the first N rows for performance. If there is larger content in cells below it will be cut off
        N = 100
        for i in range(self.dataView.model().rowCount())[:N]:
            mi = self.dataView.model().index(i, column_index)
            text = self.dataView.model().data(mi)
            w = self.dataView.fontMetrics().boundingRect(text).width()
            width = max(width, w)

        # Repeat for header cells
        for i in range(self.columnHeader.model().rowCount()):
            mi = self.columnHeader.model().index(i, column_index)
            text = self.columnHeader.model().data(mi)
            text = text[len(self.prefix) :] if text.startswith(self.prefix) else text
            w = self.columnHeader.fontMetrics().boundingRect(text).width()
            width = max(width, w)

        padding = 20
        width += padding + extra_padding

        self.columnHeader.setColumnWidth(column_index, width)
        self.dataView.setColumnWidth(column_index, width)

        self.dataView.updateGeometry()
        self.columnHeader.updateGeometry()

        return width

    def auto_size_row(self, row_index):
        """
        Set the size of row at row_index to fix its contents
        """
        height = 24

        self.indexHeader.setRowHeight(row_index, height)
        self.dataView.setRowHeight(row_index, height)

        self.dataView.updateGeometry()
        self.indexHeader.updateGeometry()

    def scroll_to_column(self, column=0):
        index = self.dataView.model().index(0, column)
        self.dataView.scrollTo(index)
        self.columnHeader.selectColumn(column)
        self.columnHeader.on_selectionChanged(force=True)

    def keyPressEvent(self, event):
        QtWidgets.QWidget.keyPressEvent(self, event)
        mods = event.modifiers()

        # Ctrl+C
        if event.key() == Qt.Key_C and (mods & Qt.ControlModifier):
            self.copy(header=True)
        # Ctrl+Shift+C
        elif (
            event.key() == Qt.Key_C
            and (mods & Qt.ShiftModifier)
            and (mods & Qt.ControlModifier)
        ):
            self.copy(header=True)
        else:
            self.dataView.keyPressEvent(event)

    def copy(self, header=False):
        """
        Copy the selected cells to clipboard in an Excel-pasteable format
        """
        # Get the bounds using the top left and bottom right selected cells

        fmt = self.dataView.model().format

        # Copy from data, columns, or index depending which has focus
        if header or self.dataView.hasFocus():
            indexes = self.dataView.selectionModel().selection().indexes()
            rows = [ix.row() for ix in indexes]
            cols = [ix.column() for ix in indexes]

            temp_df = self.pgdf.df
            df = temp_df.iloc[min(rows) : max(rows) + 1, min(cols) : max(cols) + 1]

        elif self.indexHeader.hasFocus():
            indexes = self.indexHeader.selectionModel().selection().indexes()
            rows = [ix.row() for ix in indexes]
            cols = [ix.column() for ix in indexes]

            temp_df = self.pgdf.df.index.to_frame()
            df = temp_df.iloc[min(rows) : max(rows) + 1, min(cols) : max(cols) + 1]

        elif self.columnHeader.hasFocus():
            indexes = self.columnHeader.selectionModel().selection().indexes()
            rows = [ix.row() for ix in indexes]
            cols = [ix.column() for ix in indexes]

            # Column header should be horizontal so we transpose
            temp_df = self.pgdf.df.columns.to_frame().transpose()
            df = temp_df.iloc[min(rows) : max(rows) + 1, min(cols) : max(cols) + 1]
        else:
            return

        if fmt in ("hex", "bin") and len(df):
            fmt = "{:X}" if fmt == "hex" else "{:b}"

            for name in df.columns:
                col = df[name]
                if isinstance(col.values[0], np.integer):
                    col = pd.Series([fmt.format(val) for val in col], index=df.index)
                    df[name] = col

        if self.dataView.model().float_precision != -1:
            decimals = self.dataView.model().float_precision
            for name in df.columns:
                col = df[name]
                if isinstance(col.values[0], np.floating):
                    col = col.round(decimals)
                    df[name] = col
            float_format = f"%.{decimals}f"
        else:
            float_format = "%.16f"

        # If I try to use df.to_clipboard without starting new thread, large selections give access denied error
        if df.shape == (1, 1):
            # Special case for single-cell copy, excel=False removes the trailing \n character.
            threading.Thread(
                target=lambda df: df.to_clipboard(
                    index=header,
                    header=header,
                    excel=False,
                    float_format=float_format,
                ),
                args=(df,),
            ).start()
        else:
            threading.Thread(
                target=lambda df: df.to_clipboard(
                    index=header,
                    header=header,
                    float_format=float_format,
                ),
                args=(df,),
            ).start()

    def show_column_menu(self, column_ix_or_name):
        if isinstance(self.pgdf.df.columns, pd.MultiIndex):
            logger.info("Column menu not implemented for MultiIndex")
            return

        if type(column_ix_or_name) == str:
            column_ix = list(self.pgdf.df.columns).index(column_ix_or_name)
        else:
            column_ix = column_ix_or_name

        point = QtCore.QPoint(
            self.columnHeader.columnViewportPosition(column_ix)
            + self.columnHeader.columnWidth(column_ix)
            - 15,
            self.columnHeader.geometry().bottom() - 6,
        )

        menu = ColumnMenu(self.pgdf, column_ix, self)
        menu.show_menu(self.columnHeader.mapToGlobal(point))

    def _remove_column(self, ix):
        for model in [self.dataView.model(), self.columnHeader.model()]:
            parent = QtCore.QModelIndex()
            model.beginRemoveColumns(parent, ix, ix)
            model.endRemoveColumns()

    def _move_column(self, ix, new_ix, refresh=True):
        for view in [self.dataView, self.columnHeader]:
            model = view.model()
            column_widths = [view.columnWidth(ix) for ix in range(model.columnCount())]
            column_widths.insert(new_ix, column_widths.pop(ix))

            # Set width of destination column to the width of the source column
            for j in range(len(column_widths)):
                view.setColumnWidth(j, column_widths[j])

        if refresh:
            self.refresh_ui()

    def resizeEvent(self, event):
        super().resizeEvent(event)
        self.update_horizontal_scroll()

    def refresh_ui(self):
        self.models = [
            self.dataView.model(),
            self.columnHeader.model(),
            self.indexHeader.model(),
            self.columnHeaderNames.model(),
            self.indexHeaderNames.model(),
        ]

        for model in self.models:
            model.beginResetModel()
            model.endResetModel()

        # Update sizing
        for view in [self.columnHeader, self.indexHeader, self.dataView]:
            view.updateGeometry()
